import numpy as np
import typing
import effector.bin_splitting as be


class DynamicProgramming:
    def __init__(self,
                 max_nof_bins: int = 20,
                 min_points_per_bin: int = 10,
                 discount: float = 0.3,
                 cat_limit: int = 15):
        self.max_nof_bins = max_nof_bins
        self.min_points_per_bin = min_points_per_bin
        self.discount = discount
        self.cat_limit = cat_limit


class Greedy:
    def __init__(self,
                 init_nof_bins: int = 100,
                 min_points_per_bin: int = 10,
                 discount: float = 0.3,
                 cat_limit: int = 15
                 ):
        self.max_nof_bins = init_nof_bins
        self.min_points_per_bin = min_points_per_bin
        self.discount = discount
        self.cat_limit = cat_limit


class Fixed:
    def __init__(self,
                 nof_bins: int = 100,
                 min_points_per_bin=10,
                 cat_limit: int = 15
                 ):
        self.nof_bins = nof_bins
        self.min_points_per_bin = min_points_per_bin
        self.cat_limit = cat_limit


def find_limits(
        data: np.ndarray,
        data_effect: typing.Union[None, np.ndarray],
        feature: int,
        axis_limits: np.ndarray,
        binning_method: typing.Union[str, DynamicProgramming, Greedy, Fixed]):
    """Find the limits of the bins for a specific feature.

    Parameters
    ----------
    data: numpy array, shape (n_samples, n_features)
        Data matrix
    data_effect: numpy array, shape (n_samples, n_features)
        Jacobian matrix
    feature: int
        Index of the feature of interest
    axis_limits: numpy array
        Axis limits, shape (n_features, 2)
    binning_method: str or instance of appropriate binning class
        Binning method to use

    Returns
    -------
    bin_est: instance of appropriate binning class
        Binning estimator
    """
    if isinstance(binning_method, str):
        assert binning_method in ["dp", "greedy", "fixed"]
    assert isinstance(axis_limits, np.ndarray)

    if isinstance(binning_method, Fixed):
        bin_est = be.Fixed(data, data_effect, feature=feature, axis_limits=axis_limits)
        bin_est.find(nof_bins=binning_method.nof_bins, min_points=binning_method.min_points_per_bin, cat_limit=binning_method.cat_limit)
    elif binning_method == "fixed":
        bin_est = be.Fixed(data, data_effect, feature=feature, axis_limits=axis_limits)
        bin_est.find(nof_bins=20, min_points=0, cat_limit=15)
    elif isinstance(binning_method, Greedy):
        bin_est = be.Greedy(data, data_effect, feature=feature, axis_limits=axis_limits)
        bin_est.find(
            min_points=binning_method.min_points_per_bin, init_nof_bins=binning_method.max_nof_bins, discount=binning_method.discount, cat_limit=binning_method.cat_limit
        )
    elif binning_method == "greedy":
        bin_est = be.Greedy(data, data_effect, feature=feature, axis_limits=axis_limits)
        bin_est.find(min_points=0, init_nof_bins=100, discount=0.3, cat_limit=15)
    elif isinstance(binning_method, DynamicProgramming):
        bin_est = be.DP(data, data_effect, feature=feature, axis_limits=axis_limits)
        bin_est.find(
            min_points=binning_method.min_points_per_bin, max_nof_bins=binning_method.max_nof_bins, discount=binning_method.discount, cat_limit=binning_method.cat_limit
        )
    elif binning_method == "dp":
        bin_est = be.DP(data, data_effect, feature=feature, axis_limits=axis_limits)
        bin_est.find(min_points=0, max_nof_bins=20, discount=0.3, cat_limit=15)
    return bin_est



import typing
import numpy as np
from effector.regional_effect import RegionalEffectBase
from effector import helpers
from effector.global_effect_pdp import PDP, DerivativePDP
from tqdm import tqdm

BIG_M = helpers.BIG_M


class RegionalPDPBase(RegionalEffectBase):
    def __init__(
        self,
        method_name: str,
        data: np.ndarray,
        model: callable,
        model_jac: typing.Union[None, callable] = None,
        nof_instances: typing.Union[int, str] = 100,
        axis_limits: typing.Union[None, np.ndarray] = None,
        feature_types: typing.Union[list, None] = None,
        cat_limit: typing.Union[int, None] = 10,
        feature_names: typing.Union[list, None] = None,
        target_name: typing.Union[str, None] = None,
    ):
        super(RegionalPDPBase, self).__init__(
            method_name,
            data,
            model,
            model_jac,
            None,
            nof_instances,
            axis_limits,
            feature_types,
            cat_limit,
            feature_names,
            target_name)

    def _create_heterogeneity_function(self, foi, min_points, centering, nof_instances, points_for_centering):
        def heter(data) -> float:
            if data.shape[0] < min_points:
                return BIG_M

            if self.method_name == "pdp":
                pdp = PDP(data, self.model, self.axis_limits, nof_instances=nof_instances)
            else:
                pdp = DerivativePDP(data, self.model, self.model_jac, self.axis_limits, nof_instances=nof_instances)

            try:
                pdp.fit(features=foi, centering=centering, points_for_centering=points_for_centering)
            except:
                return BIG_M

            # heterogeneity is the mean heterogeneity over the curve
            axis_limits = helpers.axis_limits_from_data(data)

            xx = np.linspace(axis_limits[:, foi][0], axis_limits[:, foi][1], 10)
            try:
                _, z = pdp.eval(feature=foi, xs=xx, heterogeneity=True)
            except:
                return BIG_M
            return np.mean(z)

        return heter

    def fit(
        self,
        features: typing.Union[int, str, list],
        heter_pcg_drop_thres: float = 0.1,
        heter_small_enough: float = 0.1,
        max_depth: int = 2,
        nof_candidate_splits_for_numerical: int = 20,
        min_points_per_subregion: int = 10,
        candidate_conditioning_features: typing.Union["str", list] = "all",
        split_categorical_features: bool = False,
        centering: typing.Union[bool, str] = False,
        nof_instances: int = "all",
        points_for_centering: int = 100,
    ):
        """
        Find the Regional PDP for a list of features.

        Args:
            features: list of features to fit
            heter_pcg_drop_thres: heterogeneity drop threshold for a split to be considered important
            heter_small_enough: heterogeneity threshold for a region to be considered homogeneous (splitting stops)
            max_depth: maximum number of splits to perform (depth of the tree)
            nof_candidate_splits_for_numerical: number of candidate splits to consider for numerical features
            min_points_per_subregion: minimum allowed number of points in a subregion (otherwise the split is not considered as valid)
            candidate_conditioning_features: list of features to consider as conditioning features for the candidate splits
            split_categorical_features
        """

        assert min_points_per_subregion >= 2, "min_points_per_subregion must be >= 2"
        features = helpers.prep_features(features, self.dim)
        for feat in tqdm(features):
            heter = self._create_heterogeneity_function(feat, min_points_per_subregion, centering, nof_instances, points_for_centering)

            self._fit_feature(
                feat,
                heter,
                heter_pcg_drop_thres,
                heter_small_enough,
                max_depth,
                nof_candidate_splits_for_numerical,
                min_points_per_subregion,
                candidate_conditioning_features,
                split_categorical_features,
            )

            # todo add methdod args


class RegionalPDP(RegionalPDPBase):
    def __init__(
        self,
        data: np.ndarray,
        model: callable,
        nof_instances: typing.Union[int, str] = 1000,
        axis_limits: typing.Union[None, np.ndarray] = None,
        feature_types: typing.Union[list, None] = None,
        cat_limit: typing.Union[int, None] = 10,
        feature_names: typing.Union[list, None] = None,
        target_name: typing.Union[str, None] = None,
    ):
        super(RegionalPDP, self).__init__(
            "pdp",
            data,
            model,
            None,
            nof_instances,
            axis_limits,
            feature_types,
            cat_limit,
            feature_names,
            target_name)


class RegionalDerivativePDP(RegionalPDPBase):
    def __init__(
        self,
        data: np.ndarray,
        model: callable,
        model_jac: typing.Union[None, callable] = None,
        nof_instances: typing.Union[int, str] = 1000,
        axis_limits: typing.Union[None, np.ndarray] = None,
        feature_types: typing.Union[list, None] = None,
        cat_limit: typing.Union[int, None] = 10,
        feature_names: typing.Union[list, None] = None,
        target_name: typing.Union[str, None] = None,
    ):
        super(RegionalDerivativePDP, self).__init__(
            "d-pdp",
            data,
            model,
            model_jac,
            nof_instances,
            axis_limits,
            feature_types,
            cat_limit,
            feature_names,
            target_name)


import scipy.integrate as integrate
import numpy as np


def integrate_1d_quad(func, start, stop):
    """Computes normalization constant of PDP of feature

    :param s: index of feature
    :returns:

    """

    def wrapper(x):
        x = np.array([x])
        return func(x)

    y = integrate.quad(wrapper, start, stop)[0]
    return y


def integrate_1d_linspace(func, start, stop):
    """

    :param s: index of feature
    :returns:

    """
    x = np.linspace(start, stop, 1000)
    x = 0.5 * (x[:-1] + x[1:])
    z = np.sum(func(x)) * (stop - start) / 1000
    return z


def mean_1d_linspace(func, start, stop, nof_points=100):
    """Computes \int_{start}^{stop} func(x) dx
    func(x) -> y: gets 1D np.array and return 1D np.array

    :param s: index of feature
    :returns:

    """
    x = np.linspace(start, stop, nof_points)
    x = 0.5 * (x[:-1] + x[1:])
    z = np.mean(func(x))
    return z


def expectation_1D(xs, func, p_xc, s, start=-np.inf, stop=np.inf):
    """

    :param func:  (N, D) -> (N)
    :param xs: np.float
    :param p_xc: p_xc float -> [0,1]
    :param s: index of feature of interest
    :param start: left limit of integral
    :param stop: right limit of integral
    :returns:

    """

    def func_wrap(xc):
        x = np.array([xs, xc]) if s == 0 else np.array([xc, xs])
        x = np.expand_dims(x, axis=0)
        return func(x)[0] * p_xc(xc)

    return integrate.quad(func_wrap, start, stop)


def expecation_2D(xs, func, p_xc, s, axis_limits):
    """

    :param func:  (N, D) -> (N)
    :param xs: np.float
    :param p_xc: p_xc (float, float)-> [0,1]
    :param s: index of feature of interest
    :param start: left limit of integral
    :param stop: right limit of integral
    :returns:

    """

    def func_wrap(xc1, xc2):
        if s == 0:
            x = np.array([xs, xc1, xc2])
        elif s == 1:
            x = np.array([xc1, xs, xc2])
        elif s == 2:
            x = np.array([xc1, xc2, xs])
        x = np.expand_dims(x, axis=0)
        return func(x)[0] * p_xc(xc1, xc2)

    def gfun(xc2):
        return axis_limits[0, 1]

    def hfun(xxc2):
        return axis_limits[1, 1]

    return integrate.dblquad(
        func_wrap, axis_limits[0, 2], axis_limits[1, 2], gfun, hfun
    )



import numpy as np
from typing import Callable, List, Optional, Union, Tuple
from effector import helpers
from abc import ABC, abstractmethod


class GlobalEffectBase(ABC):
    empty_symbol = helpers.EMPTY_SYMBOL

    def __init__(
            self,
            method_name: str,
            data: np.ndarray,
            model: Callable,
            nof_instances: Union[int, str] = 1000,
            axis_limits: Optional[np.ndarray] = None,
            avg_output: Optional[float] = None,
            feature_names: Optional[List] = None,
            target_name: Optional[str] = None,
    ) -> None:
        """
        Constructor for the FeatureEffectBase class.

        Args:
            method_name: the name of the method used to compute the feature effect
            data: the design matrix

                - shape: `(N,D)`
            model: the black-box model. Must be a `Callable` with:

                - input: `ndarray` of shape `(N, D)`
                - output: `ndarray` of shape `(N, )`

            axis_limits: The limits of the feature effect plot along each axis

                - use a `ndarray` of shape `(2, D)`, to specify them manually
                - use `None`, to be inferred from the data

            avg_output: the average output of the model on the data

                - use a `float`, to specify it manually
                - use `None`, to be inferred as `np.mean(model(data))`

            feature_names: The names of the features

                - use a `list` of `str`, to specify the name manually. For example: `["age", "weight", ...]`
                - use `None`, to keep the default names: `["x_0", "x_1", ...]`

            target_name: The name of the target variable

                - use a `str`, to specify it name manually. For example: `"price"`
                - use `None`, to keep the default name: `"y"`

        """
        assert data.ndim == 2

        self.method_name: str = method_name.lower()

        # select nof_instances from the data
        self.nof_instances, self.indices = helpers.prep_nof_instances(
            nof_instances, data.shape[0]
        )
        data = data[self.indices, :]

        self.data: np.ndarray = data
        self.dim = self.data.shape[1]

        self.model: Callable = model

        self.avg_output = (
            avg_output if avg_output is not None else np.mean(self.model(self.data))
        )

        axis_limits = (
            helpers.axis_limits_from_data(data) if axis_limits is None else axis_limits
        )
        self.axis_limits: np.ndarray = axis_limits

        self.feature_names: Union[None, list] = feature_names
        self.target_name: Union[None, str] = target_name

        # state variable
        self.is_fitted: np.ndarray = np.ones([self.dim]) < 1

        # parameters used when fitting the feature effect
        self.method_args: dict = {}

        # dictionary with all the information required for plotting or evaluating the feature effect
        self.feature_effect: dict = {}

    @abstractmethod
    def fit(self, features: Union[int, str, list] = "all", **kwargs) -> None:
        """Fit the feature effect for the given features.

        Args:
            features: the features to fit. If set to "all", all the features will be fitted.
        """
        raise NotImplementedError

    @abstractmethod
    def plot(self, feature: int, *args) -> None:
        """

        Parameters
        ----------
        feature: index of the feature to plot
        *args: all other plot-specific arguments
        """
        raise NotImplementedError

    def refit(self, feature, centering):
        """Checks if refitting is needed.
        """
        if not self.is_fitted[feature]:
            return True
        else:
            if centering is not False:
                if self.method_args["feature_" + str(feature)]["centering"] != centering:
                    return True
        return False

    @abstractmethod
    def eval(
        self,
        feature: int,
        xs: np.ndarray,
        heterogeneity: bool = False,
        centering: Union[bool, str] = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """Evaluate the effect of the s-th feature at positions `xs`.

        Notes:
            This is a common method among all the FE classes.

        Args:
            feature: index of feature of interest
            xs: the points along the s-th axis to evaluate the FE plot

              - `np.ndarray` of shape `(T, )`

            heterogeneity: whether to return the heterogeneity measures.

                  - if `heterogeneity=False`, the function returns the mean effect at the given `xs`
                  - If `heterogeneity=True`, the function returns `(y, std)` where `y` is the mean effect and `std` is the standard deviation of the mean effect

            centering: whether to center the PDP

                - If `centering` is `False`, the PDP not centered
                - If `centering` is `True` or `zero_integral`, the PDP is centered around the `y` axis.
                - If `centering` is `zero_start`, the PDP starts from `y=0`.

        Returns:
            the mean effect `y`, if `heterogeneity=False` (default) or a tuple `(y, std, estimator_var)` otherwise

        Notes:
            * If `centering` is `False`, the plot is not centered
            * If `centering` is `True` or `"zero_integral"`, the plot is centered by subtracting its mean.
            * If `centering` is `"zero_start"`, the plot starts from zero.

        Notes:
            * If `heterogeneity` is `False`, the plot returns only the mean effect `y` at the given `xs`.
            * If `heterogeneity` is `True`, the plot returns `(y, std)` where:
                * `y` is the mean effect
                * `std` is the standard deviation of the mean effect
        """
        raise NotImplementedError



import typing
import numpy as np
import itertools
import re
from effector import helpers, utils


BIG_M = helpers.BIG_M


class Regions:
    def __init__(
        self,
        feature: int,
        heter_func: callable,
        data: np.ndarray,
        data_effect: typing.Union[None, np.ndarray],
        feature_types: typing.Union[list, None],
        feature_names: typing.List[str],
        target_name: str,
        categorical_limit: int = 10,
        candidate_conditioning_features: typing.Union[None, list] = None,
        min_points_per_subregion: int = 10,
        nof_candidate_splits_for_numerical=20,
        max_split_levels=2,
        heter_pcg_drop_thres=0.1,
        heter_small_enough=0.1,
        split_categorical_features=False,
    ):
        # setters
        self.feature = feature
        self.data = data
        self.dim = self.data.shape[1]
        self.cat_limit = categorical_limit
        self.data_effect = data_effect
        self.feature_names = feature_names
        self.target_name = target_name
        self.min_points = min_points_per_subregion
        self.heter_func = heter_func
        self.nof_candidate_splits_for_numerical = nof_candidate_splits_for_numerical
        self.max_split_levels = max_split_levels
        self.heter_pcg_drop_thres = heter_pcg_drop_thres
        self.heter_small_enough = heter_small_enough
        self.split_categorical_features = split_categorical_features

        self.foi = self.feature
        self.foc = (
            [i for i in range(self.dim) if i != self.feature]
            if candidate_conditioning_features == "all"
            else candidate_conditioning_features
        )

        # on-init
        self.feature_types = (
            utils.get_feature_types(data, categorical_limit)
            if feature_types is None
            else feature_types
        )
        self.foc_types = [self.feature_types[i] for i in self.foc]

        # init method args
        self.method_args = {}

        # init splits
        self.splits: dict = {}
        self.important_splits: dict = {}

        self.splits_tree: typing.Union[Tree, None] = None
        self.important_splits_tree: typing.Union[Tree, None] = None

        # state variable
        self.split_found: bool = False
        self.important_splits_selected: bool = False

    def search_all_splits(self):
        """
        Iterate over all features of conditioning and choose the best split for each level in a greedy fashion.
        """
        if (
            self.feature_types[self.feature] == "cat"
            and not self.split_categorical_features
        ):
            self.splits = []
        else:
            assert self.max_split_levels <= len(
                self.foc
            ), "nof_levels must be smaller than len(foc)"

            # initialize heterogeneity
            heter_init = (
                self.heter_func(self.data, self.data_effect)
                if self.data_effect is not None
                else self.heter_func(self.data)
            )

            # initialize x_list, x_jac_list, splits
            x_list = [self.data]
            x_jac_list = [self.data_effect] if self.data_effect is not None else None
            splits = [
                {
                    "heterogeneity": [heter_init],
                    "weighted_heter": heter_init,
                    "nof_instances": [len(self.data)],
                    "split_i": -1,
                    "split_j": -1,
                    "foc": self.foc,
                }
            ]
            for lev in range(self.max_split_levels):
                # if any subregion has less than min_points, stop
                if any([len(x) < self.min_points for x in x_list]):
                    break

                # find optimal split
                split = self.single_level_splits(
                    x_list, x_jac_list, splits[-1]["heterogeneity"]
                )
                splits.append(split)

                # split data and data_effect based on the optimal split found above
                feat, pos, typ = split["feature"], split["position"], split["type"]

                if x_jac_list is not None:
                    x_jac_list = self.flatten_list(
                        [
                            self.split_dataset(x, x_jac, feat, pos, typ)
                            for x, x_jac in zip(x_list, x_jac_list)
                        ]
                    )

                x_list = self.flatten_list(
                    [self.split_dataset(x, None, feat, pos, typ) for x in x_list]
                )

                self.splits = splits

        # update state
        self.split_found = True
        return self.splits

    def single_level_splits(
        self,
        x_list: list,
        x_jac_list: typing.Union[list, None],
        heter_before: list,
    ):
        """Find all splits for a single level."""
        foc_types = self.foc_types
        foc = self.foc
        nof_splits = self.nof_candidate_splits_for_numerical
        heter_func = self.heter_func
        cat_limit = self.cat_limit

        data = self.data

        big_M = -BIG_M

        # weighted_heter_drop[i,j] (i index of foc and j index of split position) is
        # the accumulated heterogeneity drop if I split foc[i] at index j
        weighted_heter_drop = np.ones([len(foc), max(nof_splits, cat_limit)]) * big_M

        # weighted_heter[i,j] (i index of foc and j index of split position) is
        # the accumulated heterogeneity if I split foc[i] at index j
        weighted_heter = np.ones([len(foc), max(nof_splits, cat_limit)]) * big_M

        # list with len(foc) elements
        # each element is a list with the split positions for the corresponding feature of conditioning
        candidate_split_positions = [
            self.find_positions_cat(data, foc_i)
            if foc_types[i] == "cat"
            else self.find_positions_cont(data, foc_i, nof_splits)
            for i, foc_i in enumerate(foc)
        ]

        # exhaustive search on all split positions
        for i, foc_i in enumerate(foc):
            for j, position in enumerate(candidate_split_positions[i]):
                # split datasets
                x_list_2 = self.flatten_list(
                    [
                        self.split_dataset(x, None, foc_i, position, foc_types[i])
                        for x in x_list
                    ]
                )
                if x_jac_list is not None:
                    x_jac_list_2 = self.flatten_list(
                        [
                            self.split_dataset(x, x_jac, foc_i, position, foc_types[i])
                            for x, x_jac in zip(x_list, x_jac_list)
                        ]
                    )

                # sub_heter: list with the heterogeneity after split of foc_i at position j
                if x_jac_list is None:
                    sub_heter = [heter_func(x) for x in x_list_2]
                else:
                    sub_heter = [
                        heter_func(x, x_jac) for x, x_jac in zip(x_list_2, x_jac_list_2)
                    ]

                # heter_drop: list with the heterogeneity drop after split of foc_i at position j
                heter_drop = np.array(
                    self.flatten_list(
                        [
                            [
                                heter_bef - sub_heter[int(2 * i)],
                                heter_bef - sub_heter[int(2 * i + 1)],
                            ]
                            for i, heter_bef in enumerate(heter_before)
                        ]
                    )
                )
                # populations: list with the number of instances in each dataset after split of foc_i at position j
                populations = np.array([len(xx) for xx in x_list_2])
                # weights analogous to the populations in each split
                weights = (populations + 1) / (np.sum(populations + 1))
                # weighted_heter_drop[i,j] is the weighted accumulated heterogeneity drop if I split foc[i] at index j
                weighted_heter_drop[i, j] = np.sum(heter_drop * weights)
                # weighted_heter[i,j] is the weighted accumulated heterogeneity if I split foc[i] at index j
                weighted_heter[i, j] = np.sum(weights * np.array(sub_heter))

        # find the split with the largest weighted heterogeneity drop
        i, j = np.unravel_index(
            np.argmax(weighted_heter_drop, axis=None), weighted_heter_drop.shape
        )
        feature = foc[i]
        position = candidate_split_positions[i][j]
        split_positions = candidate_split_positions[i]

        # how many instances in each dataset after the min split
        x_list_2 = self.flatten_list(
            [
                self.split_dataset(x, None, foc[i], position, foc_types[i])
                for x in x_list
            ]
        )

        nof_instances = [len(x) for x in x_list_2]
        if x_jac_list is None:
            sub_heter = [heter_func(x) for x in x_list_2]
        else:
            x_jac_list_2 = self.flatten_list(
                [
                    self.split_dataset(x, x_jac, foc[i], position, foc_types[i])
                    for x, x_jac in zip(x_list, x_jac_list)
                ]
            )

            sub_heter = [
                heter_func(x, x_jac) for x, x_jac in zip(x_list_2, x_jac_list_2)
            ]

        split = {
            "feature": feature,
            "position": position,
            "range": [np.min(data[:, feature]), np.max(data[:, feature])],
            "candidate_split_positions": split_positions,
            "nof_instances": nof_instances,
            "type": foc_types[i],
            "heterogeneity": sub_heter,
            "split_i": i,
            "split_j": j,
            "foc": foc,
            "weighted_heter_drop": weighted_heter_drop[i, j],
            "weighted_heter": weighted_heter[i, j],
        }
        return split

    def choose_important_splits(self):
        assert self.split_found, "No splits found for feature {}".format(self.feature)

        # if split is empty, skip
        if len(self.splits) == 0:
            optimal_splits = {}
        # if initial heterogeneity is BIG_M, skip
        elif self.splits[0]["weighted_heter"] == BIG_M:
            optimal_splits = {}
        # if initial heterogeneity is small right from the beginning, skip
        elif self.splits[0]["weighted_heter"] < self.heter_small_enough:
            optimal_splits = {}
        else:
            splits = self.splits

            # accept split if heterogeneity drops over 20%
            heter = np.array([splits[i]["weighted_heter"] for i in range(len(splits))])
            heter_drop = (heter[:-1] - heter[1:]) / heter[:-1]
            split_valid = heter_drop > self.heter_pcg_drop_thres

            # if all are negative, return nothing
            if np.sum(split_valid) == 0:
                optimal_splits = {}
            # if all are positive, return all
            elif np.sum(split_valid) == len(split_valid):
                optimal_splits = splits[1:]
            else:
                # find first negative split
                first_negative = np.where(split_valid == False)[0][0]

                # if first negative is the first split, return nothing
                if first_negative == 0:
                    optimal_splits = {}
                else:
                    optimal_splits = splits[1 : first_negative + 1]

        # update state variable
        self.important_splits_selected = True
        self.important_splits = optimal_splits
        return optimal_splits

    def split_dataset(self, x, x_jac, feature, position, feat_type):
        if feat_type == "cat":
            ind_1 = x[:, feature] == position
            ind_2 = x[:, feature] != position
        else:
            ind_1 = x[:, feature] < position
            ind_2 = x[:, feature] >= position

        if x_jac is None:
            X1 = x[ind_1, :]
            X2 = x[ind_2, :]
        else:
            X1 = x_jac[ind_1, :]
            X2 = x_jac[ind_2, :]

        return X1, X2

    def find_positions_cat(self, x, feature):
        return np.unique(x[:, feature])

    def find_positions_cont(self, x, feature, nof_splits):
        step = (np.max(x[:, feature]) - np.min(x[:, feature])) / nof_splits
        return np.min(x[:, feature]) + (np.arange(nof_splits) + 0.5) * step

    def flatten_list(self, l):
        return [item for sublist in l for item in sublist]

    def splits_to_tree(self, only_important=False, scale_x_list=None):
        if len(self.splits) == 0:
            return None

        nof_instances = self.splits[0]["nof_instances"][0]
        tree = Tree()
        # format with two decimals
        data = {
            "heterogeneity": self.splits[0]["heterogeneity"][0],
            "feature_name": self.feature,
            "nof_instances": self.splits[0]["nof_instances"][0],
            "data": self.data,
            "data_effect": self.data_effect,
            "weight": 1.
        }

        feature_name = self.feature_names[self.feature]
        tree.add_node(feature_name, None, data=data, level=0)
        parent_level_nodes = [feature_name]
        parent_level_data = [self.data]
        parent_level_data_effect = [self.data_effect]
        splits = self.important_splits if only_important else self.splits[1:]

        for i, split in enumerate(splits):

            # nof nodes to add
            nodes_to_add = len(split["nof_instances"])

            new_parent_level_nodes = []
            new_parent_level_data = []
            new_parent_level_data_effect = []

            # find parent
            for j in range(nodes_to_add):
                parent_name = parent_level_nodes[int(j/2)]
                parent_data = parent_level_data[int(j/2)]
                parent_data_effect = parent_level_data_effect[int(j/2)]

                # prepare data

                foc_name = self.feature_names[split["feature"]]
                foc = split["feature"]
                pos = split["position"]
                if scale_x_list is not None:
                    mean = scale_x_list[foc]["mean"]
                    std = scale_x_list[foc]["std"]
                    pos_scaled = std * split["position"] + mean
                else:
                    pos_scaled = pos

                pos_small = pos_scaled.round(2)

                data_1, data_2 = self.split_dataset(parent_data, None, foc, pos, split["type"])
                if self.data_effect is not None:
                    data_effect_1, data_effect_2 = self.split_dataset(parent_data, parent_data_effect, foc, pos, split["type"])
                else:
                    data_effect_1, data_effect_2 = None, None

                data_new = data_1 if j % 2 == 0 else data_2
                data_effect_new = data_effect_1 if j % 2 == 0 else data_effect_2
                if j % 2 == 0:
                    if split["type"] == "cat":
                        name = foc_name + " == {}".format(pos_small)
                        comparison = "=="
                    else:
                        name = foc_name + " <= {}".format(pos_small)
                        comparison = "<="
                else:
                    if split["type"] == "cat":
                        name = foc_name + " != {}".format(pos_small)
                        comparison = "!="
                    else:
                        name = foc_name + "  > {}".format(pos_small)
                        comparison = ">"

                name = parent_name + " | " + name if nodes_to_add == 2 else parent_name + " and " + name

                data = {
                "heterogeneity": split["heterogeneity"][j],
                "weight": float(data_new.shape[0]) / nof_instances,
                "position": split["position"],
                "feature": split["feature"],
                "feature_type": split["type"],
                "range": split["range"],
                "candidate_split_positions": split["candidate_split_positions"],
                "nof_instances": split["nof_instances"][j],
                "data": data_new,
                "data_effect": data_effect_new,
                "comparison": comparison,
                }

                tree.add_node(name, parent_name=parent_name, data=data, level=i+1)

                new_parent_level_nodes.append(name)
                new_parent_level_data.append(data_new)
                new_parent_level_data_effect.append(data_effect_new)

            # update parent_level_nodes
            parent_level_nodes = new_parent_level_nodes
            parent_level_data = new_parent_level_data
            parent_level_data_effect = new_parent_level_data

        return tree


class Node:
    def __init__(self, idx, name, parent_node, data, level):
        self.idx = idx
        self.name = name
        self.parent_node = parent_node
        self.data = data
        self.level = level

        self.heterogeneity = data["heterogeneity"]
        self.weight = data["weight"]
        self.nof_instances = data["nof_instances"]

        self.foc = data["feature"] if "feature" in data else None
        self.foc_type = data["feature_type"] if "feature_type" in data else None
        self.foc_position = data["position"] if "position" in data else None
        self.comparison = data["comparison"] if "comparison" in data else None
        self.candidate_split_positions = data["candidate_split_positions"] if "candidate_split_positions" in data else None
        self.range = data["range"] if "range" in data else None

    def show(self, show_data=False):
        print("Node id: ", self.idx)
        print("name: ", self.name)
        print("parent name: ", self.parent_node.name if self.parent_node is not None else None)
        print("level: ", self.level)

        print("heterogeneity: ", self.heterogeneity)
        print("weight: ", self.weight)
        print("nof_instances: ", self.nof_instances)

        print("foc: ", self.foc)
        print("foc_type: ", self.foc_type)
        print("foc_position: ", self.foc_position)
        print("comparison: ", self.comparison)

        print("data: ", self.data) if show_data else None


class Tree:
    def __init__(self):
        self.nodes = []
        self.idx = 0

    def rename_nodes(self, scale_x_per_feature):
        nodes = self.nodes

        for node in nodes:
            node.name = self._rename_node(node.name, scale_x_per_feature)

    def _rename_node(self, name, scale):
        pattern = r"(x_\d+) (==|!=|<=|<|>=|>) ([\d\.]+)"
        for match in re.finditer(pattern, name):
            var_name = match.group(1)
            symb = match.group(2)
            orig_value = match.group(3)
            feat_i = int(var_name[2:])
            std = scale["feature_" + str(feat_i)]["std"]
            mean = scale["feature_" + str(feat_i)]["mean"]
            scaled_value = std * float(orig_value) + mean
            new_string = name.replace(orig_value, str(scaled_value))
            name = new_string
        return name

    def add_node(self, name, parent_name, data, level):
        if parent_name is None:
            parent_node = None
        else:
            assert parent_name in [node.name for node in self.nodes]
            parent_node = self.get_node(parent_name)

        idx = self.idx
        self.idx += 1
        node = Node(idx, name, parent_node, data, level)
        self.nodes.append(node)

    def get_node(self, name):
        node = None
        for node_i in self.nodes:
            if node_i.name == name:
                node = node_i
                break
        return node

    def get_level_nodes(self, level):
        nodes = []
        for node_i in self.nodes:
            if node_i.level == level:
                nodes.append(node_i)
        return nodes

    def get_root(self):
        node = None
        for node_i in self.nodes:
            if node_i.parent_node is None:
                node = node_i
                break
        assert node is not None
        return node

    def get_children(self, name):
        children = []
        for node_i in self.nodes:
            if node_i.parent_node is not None:
                if node_i.parent_node.name == name:
                    children.append(node_i)
        return children

    def get_level_stats(self, level):
        level_nodes = self.get_level_nodes(level)
        nof_instances = self.get_root().data["nof_instances"]

        w_heter = 0
        for nod in level_nodes:
            w_heter += nod.data["heterogeneity"] * nod.data["weight"]

        return {"heterogeneity": w_heter}

    def show_full_tree(self, node=None):
        if node is None:
            node = self.get_root()

        indent = node.level*2
        print("    " * indent + "Node id: %d, name: %s, heter: %.2f || nof_instances: %5d || weight: %.2f" % (node.idx, node.name, node.data['heterogeneity'], node.data['nof_instances'], node.data['weight']))
        children = self.get_children(node.name)
        for child in children:
            self.show_full_tree(child)

    def show_level_stats(self, node=None):
        max_level = max([node.level for node in self.nodes])
        prev_heter = 0
        for lev in range(max_level+1):
            level_stats = self.get_level_stats(lev)
            if lev == 0:
                print("    " * lev*2 + "Level %.d, heter: %.2f" % (lev, level_stats['heterogeneity']))
            else:
                print("    " * lev*2 + "Level %.d, heter: %.2f || heter drop: %.2f (%.2f%%)" % (lev, level_stats['heterogeneity'], prev_heter - level_stats['heterogeneity'], 100*(prev_heter - level_stats['heterogeneity'])/prev_heter))
            prev_heter = level_stats['heterogeneity']




class DataTransformer:
    def __init__(self, splits: typing.Dict):
        self.splits = splits

    def transform(self, X):
        # feat_mapping <- to how many features each feature is mapped
        feat_mapping = []
        for split in self.splits.values():
            if len(split) == 0:
                feat_mapping.append(1)
            else:
                feat_mapping.append(2 ** len(split))

        # the enhanced data, without masking
        new_data = []
        for i in range(X.shape[1]):
            new_data.append(np.repeat(X[:, i, np.newaxis], feat_mapping[i], axis=-1))
        new_data = np.concatenate(new_data, axis=-1)

        # create mask, based on splits
        mask = np.ones(new_data.shape)
        new_columns = []
        for feat in range(X.shape[1]):
            # jj = j in the enhanced dataset
            pos = int(np.sum(feat_mapping[:feat]))

            if feat_mapping[feat] == 1:
                new_columns.append("x{}".format(feat))
                continue
            else:
                feat_splits = self.splits["feat_{}".format(feat)]
                lst = [
                    list(i) for i in itertools.product([0, 1], repeat=len(feat_splits))
                ]
                for ii, bin in enumerate(lst):
                    new_name = "x{} | ".format(feat)
                    init_col_mask = np.ones(new_data.shape[0]) * True
                    for jj, b in enumerate(bin):
                        if b == 0:
                            if feat_splits[jj]["type"] == "cat":
                                init_col_mask = np.logical_and(
                                    init_col_mask,
                                    X[:, feat_splits[jj]["feature"]]
                                    == feat_splits[jj]["position"],
                                )
                                # add with two decimals
                                new_name += "x{}={:.2f} & ".format(
                                    feat_splits[jj]["feature"],
                                    feat_splits[jj]["position"],
                                )
                            else:
                                init_col_mask = np.logical_and(
                                    init_col_mask,
                                    X[:, feat_splits[jj]["feature"]]
                                    <= feat_splits[jj]["position"],
                                )
                                new_name += "x{}<={:.2f} & ".format(
                                    feat_splits[jj]["feature"],
                                    feat_splits[jj]["position"],
                                )
                        else:
                            if feat_splits[jj]["type"] == "cat":
                                init_col_mask = np.logical_and(
                                    init_col_mask,
                                    X[:, feat_splits[jj]["feature"]]
                                    != feat_splits[jj]["position"],
                                )
                                new_name += "x{}!={:.2f} & ".format(
                                    feat_splits[jj]["feature"],
                                    feat_splits[jj]["position"],
                                )
                            else:
                                init_col_mask = np.logical_and(
                                    init_col_mask,
                                    X[:, feat_splits[jj]["feature"]]
                                    > feat_splits[jj]["position"],
                                )
                                new_name += "x{}>{:.2f} & ".format(
                                    feat_splits[jj]["feature"],
                                    feat_splits[jj]["position"],
                                )
                    # current position in mask
                    mask[:, pos + ii] = init_col_mask
                    new_columns.append(new_name[:-3])
        self.mask = mask
        self.new_data = new_data * mask
        self.new_names = new_columns
        return self.new_data


def rename_features():
    pass



import typing
from typing import Callable, List, Optional, Union, Tuple
import effector.visualization as vis
import effector.helpers as helpers
from effector.global_effect import GlobalEffectBase
import numpy as np
import shap
from scipy.interpolate import UnivariateSpline


class SHAPDependence(GlobalEffectBase):
    def __init__(
            self,
            data: np.ndarray,
            model: Callable,
            axis_limits: Optional[np.ndarray] = None,
            nof_instances: Union[int, str] = 100,
            avg_output: Optional[float] = None,
            feature_names: Optional[List[str]] = None,
            target_name: Optional[str] = None,
    ):
        """
        Constructor of the SHAPDependence class.

        Definition:
            The value of a coalition of $S$ features is estimated as:
            $$
            \hat{v}(S) = {1 \over N} \sum_{i=1}^N  f(x_S \cup x_C^i) - f(x^i)
            $$
            The value of a coalition $S$ quantifies what the values $\mathbf{x}_S$ of the features in $S$ contribute to the output of the model. It
            is the average (over all instances) difference on the output between setting features in $S$ to be $x_S$, i.e., $\mathbf{x} = (\mathbf{x}_S, \mathbf{x}_C^i)$ and leaving the instance as it is, i.e., $\mathbf{x}^i = (\mathbf{x}_S^i, \mathbf{x}_C^i)$.

            The contribution of a feature $j$ added to a coalition $S$ is estimated as:
            $$
            \hat{\Delta}_{S, j} = \hat{v}(S \cup \{j\}) - \hat{v}(S)
            $$

            The SHAP value of a feature $j$ with value $x_j$ is the average contribution of feature $j$ across all possible coalitions with a weight $w_{S, j}$:

            $$
            \hat{\phi}_j(x_j) = {1 \over N} \sum_{S \subseteq \{1, \dots, D\} \setminus \{j\}} w_{S, j} \hat{\Delta}_{S, j}
            $$

            where $w_{S, j}$ assures that the contribution of feature $j$ is the same for all coalitions of the same size. For example, there are $D-1$ ways for $x_j$ to enter a coalition of $|S| = 1$ feature, so $w_{S, j} = {1 \over D (D-1)}$ for each of them. In contrast, there is only one way for $x_j$ to enter a coaltion of $|S|=0$ (to be the first specified feature), so $w_{S, j} = {1 \over D}$.

            The SHAP Dependence Plot (SHAP-DP) is a spline $\hat{f}^{SDP}_j(x_j)$ fit to the dataset $\{(x_j^i, \hat{\phi}_j(x_j^i))\}_{i=1}^N$ using the `UnivariateSpline` function from `scipy.interpolate`.

        Notes:
            * The required parameters are `data` and `model`. The rest are optional.
            * SHAP values are computed using the `shap` package, using the class `Explainer`.
            * SHAP values are centered by default, i.e., the average SHAP value is subtracted from the SHAP values.
            * More details on the SHAP values can be found in the [original paper](https://arxiv.org/abs/1705.07874) and in the book [Interpreting Machine Learning Models with SHAP](https://christophmolnar.com/books/shap/)

        Args:
            data: the design matrix

                - shape: `(N,D)`
            model: the black-box model. Must be a `Callable` with:

                - input: `ndarray` of shape `(N, D)`
                - output: `ndarray` of shape `(N,)`

            axis_limits: The limits of the feature effect plot along each axis

                - use a `ndarray` of shape `(2, D)`, to specify them manually
                - use `None`, to be inferred from the data

            nof_instances: maximum number of instances to be used for SHAP estimation.

                - use "all", for using all instances.
                - use an `int`, for using `nof_instances` instances.

            avg_output: The average output of the model.

                - use a `float`, to specify it manually
                - use `None`, to be inferred as `np.mean(model(data))`

            feature_names: The names of the features

                - use a `list` of `str`, to specify the name manually. For example: `                  ["age", "weight", ...]`
                - use `None`, to keep the default names: `["x_0", "x_1", ...]`

            target_name: The name of the target variable

                - use a `str`, to specify it name manually. For example: `"price"`
                - use `None`, to keep the default name: `"y"`
        """
        self.nof_instances, self.indices = helpers.prep_nof_instances(
            nof_instances, data.shape[0]
        )
        data = data[self.indices, :]

        super(SHAPDependence, self).__init__(
            "SHAP DP", data, model, nof_instances, axis_limits, avg_output, feature_names, target_name
        )

    def _fit_feature(
        self,
        feature: int,
        centering: typing.Union[bool, str] = False,
        points_for_centering: int = 100,
    ) -> typing.Dict:

        # drop points outside of limits
        self.data = self.data[self.data[:, feature] >= self.axis_limits[0, feature]]
        self.data = self.data[self.data[:, feature] <= self.axis_limits[1, feature]]

        # compute shap values
        data = self.data
        shap_explainer = shap.Explainer(self.model, data)
        explanation = shap_explainer(data)

        # extract x and y pais
        yy = explanation.values[:, feature]
        xx = data[:, feature]

        # make xx monotonic
        idx = np.argsort(xx)
        xx = xx[idx]
        yy = yy[idx]

        # fit spline_mean to xx, yy pairs
        spline_mean = UnivariateSpline(xx, yy)

        # fit spline_mean to the sqrt of the residuals
        yy_std = np.abs(yy - spline_mean(xx))
        spline_std = UnivariateSpline(xx, yy_std)

        # compute norm constant
        if centering == "zero_integral":
            x_norm = np.linspace(xx[0], xx[-1], points_for_centering)
            y_norm = spline_mean(x_norm)
            norm_const = np.trapz(y_norm, x_norm) / (xx[-1] - xx[0])
        elif centering == "zero_start":
            norm_const = spline_mean(xx[0])
        else:
            norm_const = helpers.EMPTY_SYMBOL

        ret_dict = {
            "spline_mean": spline_mean,
            "spline_std": spline_std,
            "xx": xx,
            "yy": yy,
            "norm_const": norm_const,
        }
        return ret_dict

    def fit(
            self,
            features: Union[int, str, List] = "all",
            centering: Union[bool, str] = False,
            points_for_centering: Union[int, str] = 100,
    ) -> None:
        """Fit the SHAP Dependence Plot to the data.

        Notes:
            The SHAP Dependence Plot (SDP) $\hat{f}^{SDP}_j(x_j)$ is a spline fit to
            the dataset $\{(x_j^i, \hat{\phi}_j(x_j^i))\}_{i=1}^N$
            using the `UnivariateSpline` function from `scipy.interpolate`.

            The SHAP standard deviation, $\hat{\sigma}^{SDP}_j(x_j)$, is a spline fit            to the absolute value of the residuals, i.e., to the dataset $\{(x_j^i, |\hat{\phi}_j(x_j^i) - \hat{f}^{SDP}_j(x_j^i)|)\}_{i=1}^N$, using the `UnivariateSpline` function from `scipy.interpolate`.

        Args:
            features: the features to fit.
                - If set to "all", all the features will be fitted.
            centering:
                - If set to False, no centering will be applied.
                - If set to "zero_integral" or True, the integral of the feature effect will be set to zero.
                - If set to "zero_mean", the mean of the feature effect will be set to zero.

            points_for_centering: number of linspaced points along the feature axis used for centering.

                - If set to `all`, all the dataset points will be used.

        Notes:
            SHAP values are by default centered, i.e., $\sum_{i=1}^N \hat{\phi}_j(x_j^i) = 0$. This does not mean that the SHAP _curve_ is centered around zero; this happens only if the $s$-th feature of the dataset instances, i.e., the set $\{x_s^i\}_{i=1}^N$ is uniformly distributed along the $s$-th axis. So, use:

            * `centering=False`, to leave the SHAP values as they are.
            * `centering=True` or `centering=zero_integral`, to center the SHAP curve around the `y` axis.
            * `centering=zero_start`, to start the SHAP curve from `y=0`.

            SHAP values are expensive to compute.
            To speed up the computation consider using a subset of the dataset
            points for computing the SHAP values and for centering the spline.
            The default values (`points_for_fitting_spline=100`
            and `points_for_centering=100`) are a moderate choice.
        """
        centering = helpers.prep_centering(centering)
        features = helpers.prep_features(features, self.dim)

        # new implementation
        for s in features:
            self.feature_effect["feature_" + str(s)] = self._fit_feature(
                s, centering, points_for_centering
            )
            self.is_fitted[s] = True
            self.method_args["feature_" + str(s)] = {
                "centering": centering,
                "points_for_centering": points_for_centering,
            }

    def eval(
        self,
        feature: int,
        xs: np.ndarray,
        heterogeneity: bool = False,
        centering: typing.Union[bool, str] = False,
    ) -> typing.Union[np.ndarray, typing.Tuple[np.ndarray, np.ndarray]]:
        """Evaluate the effect of the s-th feature at positions `xs`.

        Args:
            feature: index of feature of interest
            xs: the points along the s-th axis to evaluate the FE plot

              - `np.ndarray` of shape `(T,)`
            heterogeneity: whether to return the heterogeneity measures.

                  - if `heterogeneity=False`, the function returns the mean effect at the given `xs`
                  - If `heterogeneity=True`, the function returns `(y, std)` where `y` is the mean effect and `std` is the standard deviation of the mean effect

            centering: whether to center the plot

                - If `centering` is `False`, the SHAP curve is not centered
                - If `centering` is `True` or `zero_integral`, the SHAP curve is centered around the `y` axis.
                - If `centering` is `zero_start`, the SHAP curve starts from `y=0`.

        Returns:
            the mean effect `y`, if `heterogeneity=False` (default) or a tuple `(y, std, estimator_var)` otherwise
        """
        centering = helpers.prep_centering(centering)

        if self.refit(feature, centering):
            self.fit(features=feature, centering=centering)

        # Check if the lower bound is less than the upper bound
        assert self.axis_limits[0, feature] < self.axis_limits[1, feature]

        yy = self.feature_effect["feature_" + str(feature)]["spline_mean"](xs)

        if centering is not False:
            norm_const = self.feature_effect["feature_" + str(feature)]["norm_const"]
            yy = yy - norm_const

        if heterogeneity:
            yy_std = self.feature_effect["feature_" + str(feature)]["spline_std"](xs)
            return yy, yy_std
        else:
            return yy

    def plot(
        self,
        feature: int,
        heterogeneity: Union[bool, str] = False,
        centering: Union[bool, str] = False,
        nof_points: int = 30,
        scale_x: Optional[dict] = None,
        scale_y: Optional[dict] = None,
        nof_shap_values: Union[int, str] = "all",
        show_avg_output: bool = False,
        y_limits: Optional[List] = None,
    ) -> None:
        """
        Plot the SHAP Dependence Plot (SDP) of the s-th feature.

        Args:
            feature: index of the plotted feature
            heterogeneity: whether to output the heterogeneity of the SHAP values

                - If `heterogeneity` is `False`, no heterogeneity is plotted
                - If `heterogeneity` is `True` or `"std"`, the standard deviation of the shap values is plotted
                - If `heterogeneity` is `"shap_values"`, the shap values are scattered on top of the SHAP curve

            centering: whether to center the SDP

                - If `centering` is `False`, the SHAP curve is not centered
                - If `centering` is `True` or `zero_integral`, the SHAP curve is centered around the `y` axis.
                - If `centering` is `zero_start`, the SHAP curve starts from `y=0`.

            nof_points: number of points to evaluate the SDP plot
            scale_x: dictionary with keys "mean" and "std" for scaling the x-axis
            scale_y: dictionary with keys "mean" and "std" for scaling the y-axis
            nof_shap_values: number of shap values to show on top of the SHAP curve
            show_avg_output: whether to show the average output of the model
            y_limits: limits of the y-axis
        """
        heterogeneity = helpers.prep_confidence_interval(heterogeneity)

        x = np.linspace(
            self.axis_limits[0, feature], self.axis_limits[1, feature], nof_points
        )

        # get the SHAP curve
        y = self.eval(feature, x, heterogeneity=False, centering=centering)
        y_std = (
            self.feature_effect["feature_" + str(feature)]["spline_std"](x)
            if heterogeneity == "std" or True
            else None
        )

        # get some SHAP values
        _, ind = helpers.prep_nof_instances(nof_shap_values, self.data.shape[0])
        yy = (
            self.feature_effect["feature_" + str(feature)]["yy"][ind]
            if heterogeneity == "shap_values"
            else None
        )
        if yy is not None and centering is not False:
            yy = yy - self.feature_effect["feature_" + str(feature)]["norm_const"]
        xx = (
            self.feature_effect["feature_" + str(feature)]["xx"][ind]
            if heterogeneity == "shap_values"
            else None
        )

        if show_avg_output:
            avg_output = helpers.prep_avg_output(self.data, self.model, self.avg_output, scale_y)
        else:
            avg_output = None

        vis.plot_shap(
            x,
            y,
            xx,
            yy,
            y_std,
            feature,
            heterogeneity=heterogeneity,
            scale_x=scale_x,
            scale_y=scale_y,
            avg_output=avg_output,
            feature_names=self.feature_names,
            target_name=self.target_name,
            y_limits=y_limits
        )



import numpy as np
import effector


def test_regional():
    np.random.seed(21)

    N = 1000
    T = 1000

    # create data, model
    data = np.stack(
        [
            np.random.uniform(-1, 1, N),
            np.random.uniform(-1, 1, N),
            np.random.randint(0, 2, N)
        ],
        axis=1)

    def model(x):
        y = np.zeros_like(x[:, 0])
        ind = np.logical_and(x[:, 1] > 0, x[:, 2] == 0)
        y[ind] = 5*x[ind, 0]
        return y

    def model_jac(x):
        y = np.zeros_like(x)
        ind = np.logical_and(x[:, 1] > 0, x[:, 2] == 0)
        y[ind, 0] = 5
        return y

    xs = np.linspace(-1, 1, T)
    methods = ["pdp", "d-pdp", "ale", "rhale", "shap"]
    for method in methods:
        if method == "pdp":
            reg_eff = effector.RegionalPDP(data, model, nof_instances=1000)
        elif method == "d-pdp":
            reg_eff = effector.RegionalDerivativePDP(data, model, model_jac, nof_instances=1000)
        elif method == "ale":
            reg_eff = effector.RegionalALE(data, model, nof_instances=1000)
        elif method == "rhale":
            reg_eff = effector.RegionalRHALE(data, model, model_jac, nof_instances=1000)
        else:
            reg_eff = effector.RegionalSHAP(data, model, nof_instances=1000)

        reg_eff.fit(0,
                    heter_pcg_drop_thres=0.1,
                    nof_candidate_splits_for_numerical=11,
                    candidate_conditioning_features="all",
                    split_categorical_features=True)
        y, std = reg_eff.eval(0, 5, xs, heterogeneity=True, centering=True)
        np.allclose(y, 5*xs, atol=0.1, rtol=0.1)
        np.allclose(std, 0, atol=0.1, rtol=0.1)



import numpy as np
import effector
import effector.bin_splitting


np.random.seed(21)


class TestExample2:
    """Simple example, where ground-truth can be computed in closed-form.

    Notes
    -----
    The black-box function is

    .. math:: f(x_1, x_2) = x_1^2 + x_1^2x_2

    """

    atol = 1.0e-2
    n = 100000
    k = 100

    @staticmethod
    def generate_samples(n: int, seed: int = None) -> np.array:
        """Generate N samples

        Parameters
        ----------
        n: int
          nof samples
        seed: int or None
          seed for generating samples

        Returns
        -------
        y: ndarray, shape: [N,2]
          the samples
        """
        if seed is not None:
            np.random.seed(seed)

        x1 = np.concatenate((np.array([0]),
                             np.random.uniform(0, 1, size=int(n-2)),
                             np.array([1])))
        x2 = np.concatenate((np.array([0]),
                             np.random.uniform(0, 1, size=int(n-2)),
                             np.array([1])))
        return np.stack([x1, x2]).T

    @staticmethod
    def f(x: np.array) -> np.array:
        return 3 + 2 * x[:, 0] - 4 * x[:, 1]

    @staticmethod
    def f_der(x):
        return np.stack([np.ones(x.shape[0]) * 2, np.ones(x.shape[0]) * -4], axis=-1)

    @staticmethod
    def ale_1(x):
        return 2*x - 1

    @staticmethod
    def ale_2(x):
        return -4*x + 2

    def test_dale(self):
        # generate data
        samples = self.generate_samples(self.n)

        # prediction
        dale = effector.RHALE(data=samples, model=self.f, model_jac=self.f_der)
        fixed = effector.binning_methods.Fixed(nof_bins=self.k, min_points_per_bin=0)
        dale.fit(features="all", binning_method=fixed, centering=True)

        x = np.linspace(0, 1, 1000)
        pred = dale.eval(feature=0, xs=x, centering=True)
        gt = self.ale_1(x)
        assert np.allclose(pred, gt, atol=self.atol)

        pred = dale.eval(feature=1, xs=x, centering=True)
        gt = self.ale_2(x)
        assert np.allclose(pred, gt, atol=self.atol)

        # DALE variable-size
        dp = effector.binning_methods.DynamicProgramming(max_nof_bins=20, min_points_per_bin=10, cat_limit=1)
        dale.fit(binning_method=dp, centering=True)

        pred = dale.eval(feature=0, xs=x, centering=True)
        gt = self.ale_1(x)
        assert np.allclose(pred, gt, atol=1.e-2)

        pred = dale.eval(feature=1, xs=x, centering=True)
        gt = self.ale_2(x)
        assert np.allclose(pred, gt, atol=1.e-2)


class TestBinEstimation:
    """
    Tests only whether the solution is valid, not if it is the optimal.
    """

    @staticmethod
    def model(x, par):
        """f(x1, x2) = a + b*x1 + x1x2"""
        ind_1 = np.logical_and(x[:, 0] >= par[0]["from"], x[:, 0] < par[0]["to"])
        ind_2 = np.logical_and(x[:, 0] >= par[1]["from"], x[:, 0] < par[1]["to"])
        ind_3 = np.logical_and(x[:, 0] >= par[2]["from"], x[:, 0] < par[2]["to"])
        ind_4 = np.logical_and(x[:, 0] >= par[3]["from"], x[:, 0] <= par[3]["to"])

        y = par[0]["b"] * x[:, 0] + par[0]["a"] + x[:, 0] * x[:, 1]
        y[ind_2] = (
            par[1]["b"] * (x[ind_2, 0] - par[1]["from"])
            + par[1]["a"]
            + x[ind_2, 0] * x[ind_2, 1]
        )
        y[ind_3] = (
            par[2]["b"] * (x[ind_3, 0] - par[2]["from"])
            + par[2]["a"]
            + x[ind_3, 0] * x[ind_3, 1]
        )
        y[ind_4] = (
            par[3]["b"] * (x[ind_4, 0] - par[3]["from"])
            + par[3]["a"]
            + x[ind_4, 0] * x[ind_4, 1]
        )

        return y

    @staticmethod
    def model_jac(x, params):
        """df/dx1 = b*x1 + x2
        df/dx2 = x1
        """
        ind_1 = np.logical_and(x[:, 0] >= params[0]["from"], x[:, 0] < params[0]["to"])
        ind_2 = np.logical_and(x[:, 0] >= params[1]["from"], x[:, 0] < params[1]["to"])
        ind_3 = np.logical_and(x[:, 0] >= params[2]["from"], x[:, 0] < params[2]["to"])
        ind_4 = np.logical_and(x[:, 0] >= params[3]["from"], x[:, 0] <= params[3]["to"])

        y = np.ones_like(x)
        y[ind_1, 0] = params[0]["b"] + x[ind_1, 1]
        y[ind_2, 0] = params[1]["b"] + x[ind_2, 1]
        y[ind_3, 0] = params[2]["b"] + x[ind_3, 1]
        y[ind_4, 0] = params[3]["b"] + x[ind_4, 1]

        y[ind_1, 1] = x[ind_1, 0]
        y[ind_2, 1] = x[ind_2, 0]
        y[ind_3, 1] = x[ind_3, 0]
        y[ind_4, 1] = x[ind_4, 0]
        return y

    @staticmethod
    def generate_samples(N, noise_level):
        """x1 ~ U(0,1)
        x2 ~ N(0, noise_level)
        """
        # eps = 1e-03
        # stop = 5
        x1 = np.concatenate(
            (
                np.array([0.0]),
                np.random.uniform(0.0, 1.0, size=int(N - 2)),
                np.array([1.0]),
            )
        )

        x2 = np.random.normal(loc=0, scale=noise_level, size=(int(x1.shape[0])))
        x = np.stack((x1, x2), axis=-1)
        return x

    @staticmethod
    def create_4_data_points():
        # creates 4 points
        x1 = np.array([0.0, 0.2, 0.8, 1.0])
        x2 = np.random.normal(loc=0, scale=0.01, size=(int(x1.shape[0])))
        x = np.stack((x1, x2), axis=-1)

        y_grad = np.array([[10, 10, -10, -10], [10, 10, -10, -10]]).T
        axis_limits = np.stack([x.min(axis=0), x.max(axis=0)])
        return x, y_grad, axis_limits

    def create_many_points(self):
        params = [
            {"a": 0.0, "b": 10, "from": 0.0, "to": 0.25},
            {"a": 3.0, "b": -10.0, "from": 0.25, "to": 0.5},
            {"a": 0.0, "b": 5.0, "from": 0.5, "to": 0.75},
            {"a": 1.0, "b": -5.0, "from": 0.75, "to": 1},
        ]

        N = 1e4
        noise_level = 0
        x = self.generate_samples(N, noise_level)
        y_grad = self.model_jac(x, params)
        axis_limits = np.stack([x.min(axis=0), x.max(axis=0)])
        return x, y_grad, axis_limits

    def test_min_points_2(self):
        """Tests with min_points = 2 => two bins must be created"""
        x, y_grad, axis_limits = self.create_4_data_points()

        # test Greedy
        min_points = 2
        est = effector.bin_splitting.Greedy(
            x, y_grad, feature=0, axis_limits=axis_limits
        )
        limits_greedy = est.find(init_nof_bins=100, discount=.3, min_points=min_points, cat_limit=1)

        assert limits_greedy.size == 3
        assert np.allclose(0, limits_greedy[0])
        assert np.allclose(1.0, limits_greedy[-1])

        # test DP
        min_points = 2
        est = effector.bin_splitting.DP(x, y_grad, feature=0, axis_limits=axis_limits)
        limits_dp = est.find(max_nof_bins=10, min_points=min_points, cat_limit=1)

        assert limits_dp.size == 3
        assert np.allclose(0, limits_dp[0])
        assert 0.2 <= limits_dp[1] <= 0.8
        assert np.allclose(1.0, limits_dp[2])

    def test_min_points_3(self):
        x, y_grad, axis_limits = self.create_4_data_points()
        gt_limits = np.array([0, 1.0])

        min_points = 3
        est = effector.bin_splitting.Greedy(
            x, y_grad, feature=0, axis_limits=axis_limits
        )
        limits_greedy = est.find(init_nof_bins=100, discount=1.05, min_points=min_points, cat_limit=1)
        assert np.allclose(gt_limits, limits_greedy)

        min_points = 3
        est = effector.bin_splitting.DP(x, y_grad, feature=0, axis_limits=axis_limits)
        limits_dp = est.find(max_nof_bins=10, min_points=min_points, cat_limit=1)
        assert np.allclose(gt_limits, limits_dp)

    def test_min_points_4(self):
        x, y_grad, axis_limits = self.create_4_data_points()
        gt_limits = np.array([0, 1.0])

        min_points = 4
        est = effector.bin_splitting.Greedy(
            x, y_grad, feature=0, axis_limits=axis_limits
        )
        limits_greedy = est.find(init_nof_bins=100, discount=1.05, min_points=min_points, cat_limit=1)
        assert np.allclose(gt_limits, limits_greedy)

        min_points = 4
        est = effector.bin_splitting.DP(x, y_grad, feature=0, axis_limits=axis_limits)
        limits_dp = est.find(max_nof_bins=10, min_points=min_points, cat_limit=1)
        assert np.allclose(gt_limits, limits_dp)

    def test_min_points_5(self):
        x, y_grad, axis_limits = self.create_4_data_points()

        min_points = 5
        est = effector.bin_splitting.Greedy(
            x, y_grad, feature=0, axis_limits=axis_limits
        )
        limits_Greedy = est.find(min_points, cat_limit=1)
        assert limits_Greedy is False

        min_points = 5
        est = effector.bin_splitting.DP(x, y_grad, feature=0, axis_limits=axis_limits)
        limits_DP = est.find(min_points=min_points, cat_limit=1)
        assert limits_DP is False

    def test_many_points(self):
        x, y_grad, axis_limits = self.create_many_points()
        tol = 0.05

        # test Greedy
        min_points = 10
        est = effector.bin_splitting.Greedy(
            x, y_grad, feature=0, axis_limits=axis_limits
        )
        limits_greedy = est.find(min_points, cat_limit=1)
        assert (
            np.sum(
                np.logical_and(
                    limits_greedy >= 0 - tol,
                    limits_greedy <= 0 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_greedy >= 0.25 - tol,
                    limits_greedy <= 0.25 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_greedy >= 0.5 - tol,
                    limits_greedy <= 0.5 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_greedy >= 0.75 - tol,
                    limits_greedy <= 0.75 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_greedy >= 1.0 - tol,
                    limits_greedy <= 1.0 + tol,
                )
            )
            >= 1
        )

        # test DP
        min_points = 10
        est = effector.bin_splitting.DP(x, y_grad, feature=0, axis_limits=axis_limits)
        limits_dp = est.find(min_points, cat_limit=1)
        assert (
            np.sum(
                np.logical_and(
                    limits_dp >= 0 - tol,
                    limits_dp <= 0 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_dp >= 0.25 - tol,
                    limits_dp <= 0.25 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_dp >= 0.5 - tol,
                    limits_dp <= 0.5 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_dp >= 0.75 - tol,
                    limits_dp <= 0.75 + tol,
                )
            )
            >= 1
        )
        assert (
            np.sum(
                np.logical_and(
                    limits_dp >= 1.0 - tol,
                    limits_dp <= 1.0 + tol,
                )
            )
            >= 1
        )



import matplotlib.pyplot as plt
import numpy as np
import typing

import scipy.interpolate

from effector import helpers


def trans_affine(x, mu, std):
    return x * std + mu


def trans_scale(x, std, square=False):
    y = x * std**2 if square else x * std
    return y


def trans_bin(x, std_x, std_y, square=False):
    y = x * (std_y / std_x) ** 2 if square else x * std_y / std_x
    return y


def ale_plot(
    ale_params: dict,
    accum_effect_func: callable,
    feature: int,
    centering: bool = True,
    error: typing.Union[None, str] = None,
    scale_x: typing.Union[None, dict] = None,
    scale_y: typing.Union[None, dict] = None,
    title: typing.Union[None, str] = None,
    avg_output: typing.Union[None, float] = None,
    feature_names: typing.Union[None, list] = None,
    target_name: typing.Union[None, str] = None,
    y_limits: typing.Union[None, tuple] = None,
    dy_limits: typing.Union[None, tuple] = None
):
    """

    Parameters
    ----------
    ale_params: Dict with ale parameters
    accum_effect_func: the accumulated effect function
    feature: which feature to plot
    centering: bool, if True, the feature is centered
    error: None or in ['std', 'stderr', 'both']
     - if 'std' the accumulated standard deviation is shown
     - if 'stderr' the accumulated standard error of the mean is shown
     - if 'both', std and stderr is plot
    scale_x: None or Dict with ['std', 'mean']
    scale_y: None or Dict with ['std', 'mean']
    savefig: False or path to store figure

    """
    # assert ale_params contains needed quantities
    assert all(
        name in ale_params
        for name in ["limits", "dx", "bin_estimator_variance", "bin_effect"]
    )

    x = np.linspace(ale_params["limits"][0], ale_params["limits"][-1], 1000)
    y, std = accum_effect_func(feature, x, True, centering)

    # transform
    x = x if scale_x is None else trans_affine(x, scale_x["mean"], scale_x["std"])
    y = y if scale_y is None else trans_affine(y, scale_y["mean"], scale_y["std"])
    std = std if scale_y is None else trans_scale(std, scale_y["std"])
    limits = (
        ale_params["limits"]
        if scale_x is None
        else trans_affine(ale_params["limits"], scale_x["mean"], scale_x["std"])
    )
    dx = (
        ale_params["dx"]
        if scale_x is None
        else trans_scale(ale_params["dx"], scale_x["std"])
    )
    bin_variance = (
        ale_params["bin_variance"]
        if scale_y is None
        else trans_bin(ale_params["bin_variance"], scale_x["std"], scale_y["std"], True)
    )
    bin_effect = (
        ale_params["bin_effect"]
        if scale_y is None
        else trans_bin(ale_params["bin_effect"], scale_x["std"], scale_y["std"])
    )

    # PLOT
    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
    if title is None:
        ax1.set_title("ALE plot")
    else:
        ax1.set_title(title)

    # first subplot
    ale_curve(ax1, x, y, avg_output=avg_output)

    # second subplot
    ale_bins(ax2, bin_effect, bin_variance, limits, dx, error, dy_limits)

    ax1.set_ylabel("y") if target_name is None else ax1.set_ylabel(target_name)

    ax1.set_ylim(y_limits[0], y_limits[1]) if y_limits is not None else None

    x_name = "x_%d" % (feature + 1) if feature_names is None else feature_names[feature]
    ax2.set_xlabel(x_name)
    ax2.set_ylabel("dy/dx")

    plt.show(block=False)


def ale_curve(ax1, x, y, avg_output=None):
    ax1.plot(x, y, "b--", label="average effect")
    if avg_output is not None:
        ax1.axhline(y=avg_output, color="black", linestyle="--", label="avg output")
    ax1.legend()


def ale_bins(ax2, bin_effects, bin_variance, limits, dx, error, dy_limits):
    bin_centers = (limits[:-1] + limits[1:]) / 2
    yerr = np.sqrt(bin_variance) if error else None
    ax2.bar(
        x=bin_centers,
        height=bin_effects,
        width=dx,
        color=(0.1, 0.1, 0.1, 0.1),
        edgecolor="blue",
        yerr=yerr,
        ecolor="red",
        label="dy_dx",
    )
    ax2.legend()
    ax2.set_ylim(dy_limits[0], dy_limits[1]) if dy_limits is not None else None


def plot_pdp_ice(
    x,
    feature,
    yy,
    title,
    confidence_interval,
    y_pdp_label,
    y_ice_label,
    scale_x: typing.Union[None, dict] = None,
    scale_y: typing.Union[None, dict] = None,
    avg_output: typing.Union[None, float] = None,
    feature_names: typing.Union[None, list] = None,
    target_name: typing.Union[None, str] = None,
    is_derivative: bool = False,
    nof_ice: typing.Union[str, int] = "all",
    y_limits: typing.Union[None, tuple] = None
):

    fig, ax = plt.subplots()
    ax.set_title(title)

    y_pdp_output = np.mean(yy, axis=1)

    # choose nof_ice randomly
    if nof_ice != "all":
        if nof_ice > yy.shape[1]:
            nof_ice = yy.shape[1]

    y_ice_outputs = yy
    std = np.std(y_ice_outputs, axis=1)
    std_err = np.sqrt(np.var(y_ice_outputs, axis=1))

    # scale x-axis
    x = x if scale_x is None else trans_affine(x, scale_x["mean"], scale_x["std"])

    # scale y-axis
    if scale_y is not None:
        std = trans_scale(std, scale_y["std"], square=False)
        std_err = trans_scale(std_err, scale_y["std"], square=False)
        if is_derivative:
            y_pdp_output = trans_scale(y_pdp_output, scale_y["std"], square=False)
            y_ice_outputs = trans_scale(y_ice_outputs, scale_y["std"], square=False)
        elif not is_derivative:
            y_pdp_output = trans_affine(y_pdp_output, scale_y["mean"], scale_y["std"])
            y_ice_outputs = trans_affine(y_ice_outputs, scale_y["mean"], scale_y["std"])

    # if avg_output is not None and scale_y is not None:
    #     if not is_derivative:
    #         avg_output = trans_affine(avg_output, scale_y["mean"], scale_y["std"])
    #     elif is_derivative:
    #         avg_output = trans_scale(avg_output, scale_y["std"], square=False)

    # plot
    if confidence_interval == "std":
        std = np.std(y_ice_outputs, axis=1)
        ax.fill_between(
            x,
            y_pdp_output - std,
            y_pdp_output + std,
            color="red",
            alpha=0.4,
            label="std",
        )
    elif confidence_interval == "std_err":
        ax.fill_between(
            x,
            y_pdp_output - std_err,
            y_pdp_output + std_err,
            color="red",
            alpha=0.4,
            label="std_err",
        )
    elif confidence_interval == "ice":
        y_ice_outputs = y_ice_outputs if nof_ice == "all" else y_ice_outputs[:, np.random.choice(range(y_ice_outputs.shape[1]), size=nof_ice, replace=False)]
        ax.plot(x, y_ice_outputs[:, 0], color="red", alpha=0.1, label=y_ice_label)
        ax.plot(x, y_ice_outputs, color="red", alpha=0.1)

    ax.plot(x, y_pdp_output, "b-", label=y_pdp_label)

    if avg_output is not None:
        ax.axhline(y=avg_output, color="black", linestyle="--", label="avg output")

    feature_name = (
        "x_%d" % (feature + 1) if feature_names is None else feature_names[feature]
    )
    ax.set_xlabel(feature_name)
    if is_derivative:
        ax.set_ylabel("dy/dx")
    else:
        ax.set_ylabel("y") if target_name is None else ax.set_ylabel(target_name)
    ax.legend()
    if y_limits is not None:
        ax.set_ylim(y_limits[0], y_limits[1])

    plt.show(block=False)
    return fig, ax


def plot_shap(
    x: np.ndarray,
    y: np.ndarray,
    xx: np.ndarray,
    yy: np.ndarray,
    y_std,
    feature: int,
    heterogeneity: str,
    scale_x: typing.Union[None, dict] = None,
    scale_y: typing.Union[None, dict] = None,
    avg_output: typing.Union[None, float] = None,
    feature_names: typing.Union[None, list] = None,
    target_name: typing.Union[None, str] = None,
    y_limits: typing.Union[None, tuple] = None
):

    fig, ax = plt.subplots()
    ax.set_title("SHAP Dependence Plot")

    # scale x-axis
    x = x if scale_x is None else trans_affine(x, scale_x["mean"], scale_x["std"])
    xx = xx if scale_x is None else trans_affine(xx, scale_x["mean"], scale_x["std"])

    # scale y-axis
    if scale_y is not None:
        y_std = trans_scale(y_std, scale_y["std"], square=False)
        y = trans_affine(y, scale_y["mean"], scale_y["std"])
        yy = trans_affine(yy, scale_y["mean"], scale_y["std"])
        # if avg_output is not None:
        #     avg_output = trans_scale(avg_output, scale_y["std"], square=False)

    # plot
    if heterogeneity == "std":
        ax.fill_between(
            x,
            y - y_std,
            y + y_std,
            color="red",
            alpha=0.4,
            label="std",
        )
    elif heterogeneity == "shap_values":
        ax.plot(xx[0], yy[0], "rx", alpha=0.5, label="SHAP values")
        ax.plot(xx, yy, "rx", alpha=0.5)

    ax.plot(x, y, "b-", label="SHAP curve")

    if avg_output is not None:
        ax.axhline(y=avg_output, color="black", linestyle="--", label="avg output")

    feature_name = (
        "x_%d" % (feature + 1) if feature_names is None else feature_names[feature]
    )
    ax.set_xlabel(feature_name)
    ax.set_ylabel("y") if target_name is None else ax.set_ylabel(target_name)
    ax.legend()
    if y_limits is not None:
        ax.set_ylim(y_limits[0], y_limits[1])
    plt.show(block=False)



import typing
from effector.regional_effect import RegionalEffectBase
from effector import helpers
import numpy as np
from effector.global_effect_ale import ALE, RHALE
from tqdm import tqdm
from effector import binning_methods
import copy
from typing import Callable, Optional, Union, List


BIG_M = helpers.BIG_M


class RegionalRHALE(RegionalEffectBase):
    def __init__(
        self,
        data: np.ndarray,
        model: Callable,
        model_jac: Optional[Callable] = None,
        instance_effects: Optional[np.ndarray] = None,
        nof_instances: Union[int, str] = "all",
        axis_limits: Optional[np.ndarray] = None,
        feature_types: Optional[List] = None,
        cat_limit: Optional[int] = 10,
        feature_names: Optional[List] = None,
        target_name: Optional[str] = None,
    ):
        """
        Regional RHALE constructor.

        Args:
            data: X matrix (N,D).
            model: the black-box model (N,D) -> (N, )
            model_jac: the black-box model Jacobian (N,D) -> (N,D)
            axis_limits: axis limits for the FE plot [2, D] or None. If None, axis limits are computed from the data.
            feature_types: list of feature types (categorical or numerical)
            cat_limit: the minimum number of unique values for a feature to be considered categorical
            feature_names: list of feature names
        """

        if instance_effects is None:
            instance_effects = model_jac(data)

        super(RegionalRHALE, self).__init__(
            "rhale",
            data,
            model,
            model_jac,
            instance_effects,
            nof_instances,
            axis_limits,
            feature_types,
            cat_limit,
            feature_names,
            target_name
        )

    def _create_heterogeneity_function(self, foi, binning_method, min_points, centering):
        binning_method = prep_binning_method(binning_method)

        def heter(data, instance_effects=None) -> float:
            if data.shape[0] < min_points:
                return BIG_M

            rhale = RHALE(data, self.model, self.model_jac, "all", None, instance_effects)
            try:
                rhale.fit(features=foi, binning_method=binning_method, centering=centering)
            except:
                return BIG_M

            # heterogeneity is the accumulated std at the end of the curve
            axis_limits = helpers.axis_limits_from_data(data)
            stop = np.array([axis_limits[:, foi][1]])
            _, z = rhale.eval(feature=foi, xs=stop, heterogeneity=True)
            return z.item()

        return heter

    def fit(
        self,
        features: typing.Union[int, str, list],
        heter_pcg_drop_thres: float = 0.1,
        heter_small_enough: float = 0.1,
        max_depth: int = 2,
        nof_candidate_splits_for_numerical: int = 20,
        min_points_per_subregion: int = 10,
        candidate_conditioning_features: typing.Union["str", list] = "all",
        split_categorical_features: bool = False,
        binning_method: typing.Union[
                str,
                binning_methods.Fixed,
                binning_methods.DynamicProgramming,
                binning_methods.Greedy,
        ] = "greedy",
        centering: typing.Union[bool, str] = False,
    ):
        """
        Find the Regional RHALE for a list of features.

        Args:
            features: list of features to fit
            heter_pcg_drop_thres: heterogeneity drop threshold for a split to be considered important
            heter_small_enough: heterogeneity threshold for a region to be considered homogeneous (splitting stops)
            binning_method: binning method to use
            max_depth: maximum number of splits to perform (depth of the tree)
            nof_candidate_splits_for_numerical: number of candidate splits to consider for numerical features
            min_points_per_subregion: minimum allowed number of points in a subregion (otherwise the split is not considered as valid)
            candidate_conditioning_features: list of features to consider as conditioning features for the candidate splits
        """

        assert min_points_per_subregion >= 2, "min_points_per_subregion must be >= 2"
        features = helpers.prep_features(features, self.dim)
        for feat in tqdm(features):
            heter = self._create_heterogeneity_function(
                feat, binning_method, min_points_per_subregion, centering
            )

            self._fit_feature(
                feat,
                heter,
                heter_pcg_drop_thres,
                heter_small_enough,
                max_depth,
                nof_candidate_splits_for_numerical,
                min_points_per_subregion,
                candidate_conditioning_features,
                split_categorical_features,
            )

            self.method_args["feature_" + str(feat)] = {
                "heter_pcg_drop_thres": heter_pcg_drop_thres,
                "heter_small_enough": heter_small_enough,
                "max_depth": max_depth,
                "nof_candidate_splits_for_numerical": nof_candidate_splits_for_numerical,
                "min_points_per_subregion": min_points_per_subregion,
                "candidate_conditioning_features": candidate_conditioning_features,
                "split_categorical_features": split_categorical_features,
                "binning_method": binning_method,
            }

    def plot(self,
             feature,
             node_idx,
             heterogeneity=False,
             centering=False,
             scale_x_list=None,
             scale_y=None,
             y_limits=None,
             dy_limits=None):

        # get data from the node
        self.refit(feature)

        if scale_x_list is not None:
            self.tree_full_scaled["feature_{}".format(feature)] = self.partitioners["feature_{}".format(feature)].splits_to_tree(False, scale_x_list)
            self.tree_pruned_scaled["feature_{}".format(feature)] = self.partitioners["feature_{}".format(feature)].splits_to_tree(True, scale_x_list)

        data, data_effect, name = self.get_node_info(feature, node_idx)
        feature_names = copy.deepcopy(self.feature_names)
        feature_names[feature] = name

        # define the method and fit
        self.method_args["feature_" + str(feature)]["heterogeneity"] = heterogeneity
        rhale = RHALE(data, self.model, self.model_jac, self.nof_instances, None, data_effect, feature_names=feature_names)
        binning_method = prep_binning_method(self.method_args["feature_" + str(feature)]["binning_method"])
        rhale.fit(features=feature, binning_method=binning_method, centering=centering)
        scale_x = scale_x_list[feature] if scale_x_list is not None else None
        rhale.plot(feature=feature, heterogeneity=heterogeneity, centering=centering, scale_x=scale_x, scale_y=scale_y, y_limits=y_limits, dy_limits=dy_limits)


class RegionalALE(RegionalEffectBase):
    def __init__(
        self,
        data: np.ndarray,
        model: callable,
        nof_instances: typing.Union[int, str] = "all",
        axis_limits: typing.Union[None, np.ndarray] = None,
        feature_types: typing.Union[list, None] = None,
        cat_limit: typing.Union[int, None] = 10,
        feature_names: typing.Union[list, None] = None,
        target_name: typing.Union[str, None] = None,
    ):
        """
        Regional RHALE constructor.

        Args:
            data: X matrix (N,D).
            model: the black-box model (N,D) -> (N, )
            model_jac: the black-box model Jacobian (N,D) -> (N,D)
            axis_limits: axis limits for the FE plot [2, D] or None. If None, axis limits are computed from the data.
            feature_types: list of feature types (categorical or numerical)
            cat_limit: the minimum number of unique values for a feature to be considered categorical
            feature_names: list of feature names
        """
        super(RegionalALE, self).__init__(
            "ale",
            data,
            model,
            None,
            None,
            nof_instances,
            axis_limits,
            feature_types,
            cat_limit,
            feature_names,
            target_name
        )

    def _create_heterogeneity_function(self, foi, binning_method, min_points, centering):
        binning_method = prep_binning_method(binning_method)
        isinstance(binning_method, binning_methods.Fixed)
        
        def heter(data, instance_effects=None) -> float:
            if data.shape[0] < min_points:
                return BIG_M

            ale = ALE(data, self.model, "all", None, instance_effects)
            try:
                ale.fit(features=foi, binning_method=binning_method, centering=centering)
            except:
                return BIG_M

            # heterogeneity is the accumulated std at the end of the curve
            axis_limits = helpers.axis_limits_from_data(data)
            stop = np.array([axis_limits[:, foi][1]])
            _, z = ale.eval(feature=foi, xs=stop, heterogeneity=True)
            return z.item()

        return heter

    def fit(
        self,
        features: typing.Union[int, str, list],
        heter_pcg_drop_thres: float = 0.1,
        heter_small_enough: float = 0.1,
        max_depth: int = 2,
        nof_candidate_splits_for_numerical: int = 20,
        min_points_per_subregion: int = 10,
        candidate_conditioning_features: typing.Union["str", list] = "all",
        split_categorical_features: bool = False,
        binning_method: typing.Union[str, binning_methods.Fixed] = binning_methods.Fixed(nof_bins=20, min_points_per_bin=0),
        centering: typing.Union[bool, str] = False,
    ):
        """
        Find the Regional RHALE for a list of features.

        Args:
            features: list of features to fit
            heter_pcg_drop_thres: heterogeneity drop threshold for a split to be considered important
            heter_small_enough: heterogeneity threshold for a region to be considered homogeneous (splitting stops)
            binning_method: binning method to use
            max_depth: maximum number of splits to perform (depth of the tree)
            nof_candidate_splits_for_numerical: number of candidate splits to consider for numerical features
            min_points_per_subregion: minimum allowed number of points in a subregion (otherwise the split is not considered as valid)
            candidate_conditioning_features: list of features to consider as conditioning features for the candidate splits
        """

        assert min_points_per_subregion >= 2, "min_points_per_subregion must be >= 2"
        features = helpers.prep_features(features, self.dim)
        for feat in tqdm(features):
            heter = self._create_heterogeneity_function(
                feat, binning_method, min_points_per_subregion, centering
            )

            self._fit_feature(
                feat,
                heter,
                heter_pcg_drop_thres,
                heter_small_enough,
                max_depth,
                nof_candidate_splits_for_numerical,
                min_points_per_subregion,
                candidate_conditioning_features,
                split_categorical_features,
            )

            self.method_args["feature_" + str(feat)] = {
                "heter_pcg_drop_thres": heter_pcg_drop_thres,
                "heter_small_enough": heter_small_enough,
                "max_depth": max_depth,
                "nof_candidate_splits_for_numerical": nof_candidate_splits_for_numerical,
                "min_points_per_subregion": min_points_per_subregion,
                "candidate_conditioning_features": candidate_conditioning_features,
                "split_categorical_features": split_categorical_features,
                "binning_method": binning_method,
                "centering": centering,
            }

    def plot(self,
             feature,
             node_idx,
             heterogeneity=False,
             centering=False,
             scale_x_list=None,
             scale_y=None,
             y_limits=None,
             dy_limits=None):

        # get data from the node
        self.refit(feature)

        if scale_x_list is not None:
            self.tree_full_scaled["feature_{}".format(feature)] = self.partitioners["feature_{}".format(feature)].splits_to_tree(False, scale_x_list)
            self.tree_pruned_scaled["feature_{}".format(feature)] = self.partitioners["feature_{}".format(feature)].splits_to_tree(True, scale_x_list)

        data, data_effect, name = self.get_node_info(feature, node_idx)
        feature_names = copy.deepcopy(self.feature_names)
        feature_names[feature] = name

        # define the method and fit
        self.method_args["feature_" + str(feature)]["heterogeneity"] = heterogeneity
        rhale = RHALE(data, self.model, self.model_jac, self.nof_instances, None, data_effect, feature_names=feature_names)
        binning_method = prep_binning_method(self.method_args["feature_" + str(feature)]["binning_method"])
        rhale.fit(features=feature, binning_method=binning_method, centering=centering)
        scale_x = scale_x_list[feature] if scale_x_list is not None else None
        rhale.plot(feature=feature, heterogeneity=heterogeneity, centering=centering, scale_x=scale_x, scale_y=scale_y, y_limits=y_limits, dy_limits=dy_limits)


def prep_binning_method(method):
    assert method in ["greedy", "dp", "fixed"] or isinstance(method, binning_methods.Fixed) or isinstance(method, binning_methods.DynamicProgramming) or isinstance(method, binning_methods.Greedy)

    if method == "greedy":
        return binning_methods.Greedy()
    elif method == "dp":
        return binning_methods.DynamicProgramming()
    elif method == "fixed":
        return binning_methods.Fixed()
    else:
        return method



import typing
from effector.regional_effect import RegionalEffectBase
from effector import helpers
import numpy as np
from effector.global_effect_shap import SHAPDependence
from tqdm import tqdm
from typing import Callable, Optional, Union, List


class RegionalSHAP(RegionalEffectBase):
    big_m = helpers.BIG_M

    def __init__(
        self,
        data: np.ndarray,
        model: Callable,
        axis_limits: Optional[np.ndarray] = None,
        nof_instances: Union[int, str] = 100,
        feature_types: Optional[List[str]] = None,
        cat_limit: Optional[int] = 10,
        feature_names: Optional[List[str]] = None,
        target_name: Optional[str] = None,
    ):
        """
        Regional SHAP constructor.

        Args:
            data: the design matrix

                - shape: `(N,D)`
            model: the black-box model. Must be a `Callable` with:

                - input: `ndarray` of shape `(N, D)`
                - output: `ndarray` of shape `(N, )`

            axis_limits: The limits of the feature effect plot along each axis

                - use a `ndarray` of shape `(2, D)`, to specify them manually
                - use `None`, to be inferred from the data

            nof_instances: maximum number of instances to be used for PDP.

                - use "all", for using all instances.
                - use an `int`, for using `nof_instances` instances.

            feature_types: The feature types.

                - use `None`, to infer them from the data; whether a feature is categorical or numerical is inferred
                from whether it exceeds the `cat_limit` unique values.
                - use a list with elements `"cat"` or `"numerical"`, to specify them manually.

            cat_limit: the minimum number of unique values for a feature to be considered categorical

            feature_names: The names of the features

                - use a `list` of `str`, to specify the name manually. For example: `                  ["age", "weight", ...]`
                - use `None`, to keep the default names: `["x_0", "x_1", ...]`

            target_name: The name of the target variable

                - use a `str`, to specify it name manually. For example: `"price"`
                - use `None`, to keep the default name: `"y"`
        """
        super(RegionalSHAP, self).__init__(
            "shap",
            data,
            model,
            None,
            None,
            nof_instances,
            axis_limits,
            feature_types,
            cat_limit,
            feature_names,
            target_name
        )

    def _create_heterogeneity_function(self, foi, min_points, centering, points_for_centering):

        def heterogeneity_function(data) -> float:
            if data.shape[0] < min_points:
                return self.big_m

            axis_limits = helpers.axis_limits_from_data(data)
            xx = np.linspace(axis_limits[:, foi][0], axis_limits[:, foi][1], 10)

            shap = SHAPDependence(data, self.model, None, self.nof_instances)
            shap.fit(foi, centering, points_for_centering)
            _, z = shap.eval(foi, xx, heterogeneity=True)
            return np.mean(z)

        return heterogeneity_function

    def fit(
            self,
            features: typing.Union[int, str, list],
            heter_pcg_drop_thres: float = 0.1,
            heter_small_enough: float = 0.1,
            max_depth: int = 2,
            nof_candidate_splits_for_numerical: int = 20,
            min_points_per_subregion: int = 10,
            candidate_conditioning_features: typing.Union["str", list] = "all",
            split_categorical_features: bool = False,
            centering: typing.Union[bool, str] = False,
            points_for_centering: int = 100,
    ):
        """
        Fit the regional SHAP.

        Args:
            features: the features to fit.
                - If set to "all", all the features will be fitted.

            heter_pcg_drop_thres: threshold for the percentage drop in heterogeneity to consider a split valid
            heter_small_enough: heterogeneity threshold for a region to be considered homogeneous (splitting stops)
            max_depth: maximum number of splits to perform (depth of the tree)
            nof_candidate_splits_for_numerical: number of candidate splits to consider for numerical features
            min_points_per_subregion: minimum allowed number of points in a subregion (otherwise the split is not considered as valid)
            candidate_conditioning_features: list of features to consider as conditioning features for the candidate splits
            split_categorical_features: whether to search for subregions in categorical features
            centering: whether to center the SHAP dependence plots before estimating the heterogeneity
            points_for_centering: number of points to use for centering
        """
        assert min_points_per_subregion >= 2, "min_points_per_subregion must be >= 2"
        features = helpers.prep_features(features, self.dim)
        for feat in tqdm(features):
            heter = self._create_heterogeneity_function(
                feat, min_points_per_subregion, centering, points_for_centering
            )

            self._fit_feature(
                feat,
                heter,
                heter_pcg_drop_thres,
                heter_small_enough,
                max_depth,
                nof_candidate_splits_for_numerical,
                min_points_per_subregion,
                candidate_conditioning_features,
                split_categorical_features,
            )

            self.method_args["feature_" + str(feat)] = {
                "heter_pcg_drop_thres": heter_pcg_drop_thres,
                "heter_small_enough": heter_small_enough,
                "max_depth": max_depth,
                "nof_candidate_splits_for_numerical": nof_candidate_splits_for_numerical,
                "min_points_per_subregion": min_points_per_subregion,
                "candidate_conditioning_features": candidate_conditioning_features,
                "split_categorical_features": split_categorical_features,
                "centering": centering,
                "points_for_centering": points_for_centering,
            }



import numpy as np
import effector


def test_pdp_1d_vectorized():
    """
    Test the vectorized version of the PDP function for a square model
    """

    np.random.seed(21)

    N = 10
    T = 100

    data = np.stack([
        np.random.rand(N + 1),
        np.linspace(0, 1, N+1)
    ], axis=1)

    model = lambda x: x[:, 1] ** 2
    model_jac = lambda x: (
        np.stack(
            [
            np.zeros_like(x[:, 1]),
            2 * x[:, 1]
            ],
        axis=1
    ))

    x = np.linspace(0, 1, T)

    # test the finite difference version
    data_effect = effector.global_effect_pdp.pdp_1d_vectorized(
        model,
        data,
        x,
        feature=1,
        heterogeneity=False,
        model_returns_jac=False,
        return_all=False,
        ask_for_derivatives=True)

    diff = np.abs(data_effect - 2 * x)
    assert np.all(diff < 1e-6)

    # test the jacobian version
    data_effect = effector.global_effect_pdp.pdp_1d_vectorized(
        model_jac,
        data,
        x,
        feature=1,
        heterogeneity=False,
        model_returns_jac=True,
        return_all=False,
        ask_for_derivatives=True)

    diff = np.abs(data_effect - 2 * x)
    assert np.all(diff < 1e-6)

    # test the pdp version
    data_effect = effector.global_effect_pdp.pdp_1d_vectorized(
        model,
        data,
        x,
        feature=1,
        heterogeneity=False,
        model_returns_jac=False,
        return_all=False,
        ask_for_derivatives=False)

    diff = np.abs(data_effect - x**2)
    assert np.all(diff < 1e-6)





import typing

import numpy as np
import effector.utils as utils
import effector.helpers as helpers
import matplotlib.pyplot as plt


class BinBase:
    big_M = helpers.BIG_M

    def __init__(self,
                 feature: int,
                 data: typing.Union[None, np.ndarray],
                 data_effect: typing.Union[None, np.ndarray],
                 axis_limits: typing.Union[None, np.ndarray]):
        """Initializer.

        Parameters
        ----------
        feature: feature index
        data: np.ndarray with X if binning is based on data, else, None
        data_effect: np.ndarray with dy/dX if binning is based on data, else, None
        axis_limits: np.ndarray (2, D) or None, if None, then axis_limits are set to [xs_min, xs_max]
        """
        # set immediately
        self.feature: int = feature
        self.xs_min: float = axis_limits[0, feature] if axis_limits is not None else data[:, feature].min()
        self.xs_max: float = axis_limits[1, feature] if axis_limits is not None else data[:, feature].max()
        self.data: np.ndarray = data
        self.data_effect: np.ndarray = data_effect

        # arguments passed to method find
        self.method_args: typing.Dict[str, typing.Any] = {}

        # set in method find
        self.method_outputs: typing.Dict[str, typing.Any] = {}

        # limits: None if not set, False if no binning is possible, np.ndarray (N, 2) if binning is possible
        self.limits: typing.Union[None, False, np.ndarray] = None

    def _bin_cost(self, start, stop, discount):
        min_points = self.method_args["min_points"]
        nof_points = self.data.shape[0]
        data = self.data[:, self.feature]
        data_effect = self.data_effect[:, self.feature]
        data, data_effect = utils.filter_points_in_bin(
            data, data_effect, np.array([start, stop])
        )

        # compute cost
        if data_effect.size < min_points:
            cost = self.big_M
        else:
            discount_for_more_points = 1 - discount * (data_effect.size / nof_points)
            cost = np.var(data_effect) * (stop - start) * discount_for_more_points
        return cost

    def _bin_valid(self, start, stop):
        """Check if creating a bin with limits [start, stop] is valid.

        Returns:
            Boolean, True if the bin is valid, False otherwise
        """
        min_points = self.method_args["min_points"]
        if min_points is None:
            return True

        xs = self.data[:, self.feature]
        # dy_dxs = self.data_effect[:, self.feature]
        filtered_points, _ = utils.filter_points_in_bin(xs, None, np.array([start, stop]))
        valid = filtered_points.size >= min_points
        return valid
        
    def _none_valid_binning(self):
        """Check if it is impossible to create any binning

        Returns:
            Boolean, True if it is impossible to create any binning, False otherwise
        """
        # if there is only one unique value, then it is impossible to create any binning
        cond_1 = len(np.unique(self.data[:, self.feature])) == 1

        # if there are less than min_points, then it is impossible to create any binning
        cond_2 = self.data[:, self.feature].size < self.method_args["min_points"]

        # if either is true, then it is impossible to create any binning
        return cond_1 or cond_2

    def _only_one_bin_possible(self):
        """Check if the only possible binning is all points in one bin

        Returns:
            Boolean, True if the only possible binning is all points in one bin, False otherwise
        """
        min_points = self.method_args["min_points"]
        # if xs is categorical, then only one bin is possible
        is_categorical = np.allclose(self.xs_min, self.xs_max)

        # if xs is not categorical, then one bin is possible if there are enough points only in one bin
        dy_dxs = self.data_effect[:, self.feature]
        enough_for_one_bin = min_points <= dy_dxs.size < 2 * min_points
        return is_categorical or enough_for_one_bin

    def _is_categorical(self, cat_limit):
        """Check if the feature is categorical, i.e. has less than cat_limit unique values, and if so, set the limits

        Returns:
            Boolean, True if the feature is categorical, False otherwise
        """
        # if unique values are leq 10, then it is categorical
        is_cat = len(np.unique(self.data[:, self.feature])) <= cat_limit
        # if only one unique value, then it is categorical and set the limits
        if len(np.unique(self.data[:, self.feature])) == 1:
            self.limits = False

        if is_cat:
            # set unique values as the center of the bins
            uniq = np.sort(np.unique(self.data[:, self.feature]))
            dx = [uniq[i + 1] - uniq[i] for i in range(len(uniq) - 1)]
            lims = np.array([uniq[0] - dx[0] / 2] + [uniq[i] + dx[i] / 2 for i in range(len(uniq) - 1)] + [uniq[-1] + dx[-1] / 2])

            # if all limits are valid, then set them
            if np.all([self._bin_valid(lims[i], lims[i + 1]) for i in range(len(lims) - 1)]):
                self.limits = lims
            else:
                self.limits = False
        return is_cat

    def find(self, *args):
        """Find the optimal binning for the feature.
        """
        return NotImplementedError

    def plot(self, feature=0, block=False):
        assert self.limits is not None
        assert self.data is not None
        assert self.data_effect is not None

        limits = self.limits

        plt.figure()
        plt.title("Bin splitting for feature %d" % (feature + 1))
        xs = self.data[:, self.feature]
        dy_dxs = self.data_effect[:, self.feature]
        plt.plot(xs, dy_dxs, "bo", label="local effects")
        y_min = np.min(dy_dxs)
        y_max = np.max(dy_dxs)
        plt.vlines(limits, ymin=y_min, ymax=y_max, linestyles="dashed", label="bins")
        plt.xlabel("x_%d" % (feature + 1))
        plt.ylabel("dy/dx_%d" % (feature + 1))
        plt.legend()
        plt.show(block=block)


class Greedy(BinBase):
    """
    Greedy binning algorithm
    """
    def __init__(self, 
                 data: np.ndarray, 
                 data_effect: np.ndarray, 
                 feature: int, 
                 axis_limits: typing.Union[None, np.ndarray]):
        super(Greedy, self).__init__(feature, data, data_effect, axis_limits)

    def find(
        self,
        init_nof_bins: int = 100,
        discount: float = 0.3,
        min_points: typing.Union[None, int] = 10,
        cat_limit: int = 10,
    ) -> typing.Union[np.ndarray, bool]:

        self.method_args = {
            "init_nof_bins": init_nof_bins,
            "discount": discount,
            "min_points": min_points,
            "cat_limit": cat_limit,
        }


        xs_min = self.xs_min
        xs_max = self.xs_max

        if self._none_valid_binning():
            self.limits = False
        elif self._is_categorical(cat_limit):
            return self.limits
        elif self._only_one_bin_possible():
            self.limits = np.array([self.xs_min, self.xs_max])
        else:
            # limits with high resolution
            limits, _ = np.linspace(xs_min, xs_max, num=init_nof_bins + 1, endpoint=True, retstep=True)
            
            # merging
            i = 0
            merged_limits = [limits[0]]
            while i < init_nof_bins:
                # left limit is the last item of the merged_limits list
                left_lim = merged_limits[-1]

                # choose whether to close the bin
                if i == init_nof_bins - 1:
                    # if last bin, close it
                    close_bin = True
                else:
                    # bin_1, the bin if I close it here
                    bin_1_loss = self._bin_cost(left_lim, limits[i + 1], discount)
                    bin_1_valid = self._bin_valid(left_lim, limits[i + 1])

                    # bin_2: the bin if I close it in the next limit
                    bin_2_loss = self._bin_cost(left_lim, limits[i + 2], discount)
                    bin_2_valid = self._bin_valid(left_lim, limits[i + 2])

                    # if both bins valid
                    if bin_1_valid and bin_2_valid:
                        # if first zero, second positive -> close
                        if bin_1_loss == 0.0 and bin_2_loss > 0:
                            close_bin = True
                        # if both zero, keep it (we could close as well)
                        elif bin_1_loss == 0.0 and bin_2_loss == 0:
                            close_bin = False
                        # if both positive, compare and decide
                        else:
                            close_bin = False if bin_2_loss <= bin_1_loss else True
                    else:
                        # if either invalid, keep open
                        close_bin = False
                
                # if close_bin, then add the next limit to the merged_limits
                if close_bin:
                    merged_limits.append(limits[i + 1])
                
                i += 1

            # if last bin is without enough points, merge it with the previous
            if not self._bin_valid(merged_limits[-2], merged_limits[-1]):
                merged_limits = merged_limits[:-2] + merged_limits[-1:]

            # store result
            self.limits = np.array(merged_limits)
            self.method_outputs = {"limits": self.limits}
        return self.limits


class DP(BinBase):

    def __init__(self, data, data_effect, feature, axis_limits):
        super(DP, self).__init__(feature, data, data_effect, axis_limits)

    def _index_to_position(self, index_start, index_stop, K):
        dx = (self.xs_max - self.xs_min) / K
        start = self.xs_min + index_start * dx
        stop = self.xs_min + index_stop * dx
        return start, stop

    def _cost_of_move(self, index_before, index_next, K, discount):
        """Compute the cost of move.

        Computes the cost for moving from the index of the previous bin (index_before)
        to the index of the next bin (index_next).
        """

        big_M = self.big_M
        if index_before > index_next:
            cost = big_M
        elif index_before == index_next:
            cost = 0
        else:
            start, stop = self._index_to_position(index_before, index_next, K)
            cost = self._bin_cost(start, stop, discount)
        return cost

    def _argmatrix_to_limits(self, K):
        assert "argmatrix" in self.method_outputs, "argmatrix not found in method_outputs"
        argmatrix = self.method_outputs["argmatrix"]
        dx = (self.xs_max - self.xs_min) / K

        lim_indices = [int(argmatrix[-1, -1])]
        for j in range(K - 2, 0, -1):
            lim_indices.append(int(argmatrix[int(lim_indices[-1]), j]))
        lim_indices.reverse()

        lim_indices.insert(0, 0)
        lim_indices.append(argmatrix.shape[-1])

        # remove identical bins
        lim_indices_1 = []
        before = np.nan
        for i, lim in enumerate(lim_indices):
            if before != lim:
                lim_indices_1.append(lim)
                before = lim

        limits = self.xs_min + np.array(lim_indices_1) * dx
        dx_list = np.array(
            [limits[i + 1] - limits[i] for i in range(limits.shape[0] - 1)]
        )
        return limits, dx_list

    def find(self, max_nof_bins: int = 30, min_points: int = 10, discount: float = 0.3, cat_limit: int = 10):
        """Find the optimal binning."""

        self.method_args = {"max_nof_bins": max_nof_bins, "min_points": min_points, "discount": discount, "cat_limit": cat_limit}

        self.min_points = min_points
        big_M = self.big_M
        nof_limits = max_nof_bins + 1
        nof_bins = max_nof_bins

        # if is categorical, then only one bin is possible
        if self._none_valid_binning():
            self.limits = False
        elif self._is_categorical(cat_limit):
            return self.limits
        elif self._only_one_bin_possible():
            self.limits = np.array([self.xs_min, self.xs_max])
            self.dx_list = np.array([self.xs_max - self.xs_min])
        elif max_nof_bins == 1:
            self.limits = np.array([self.xs_min, self.xs_max])
            self.dx_list = np.array([self.xs_max - self.xs_min])
        else:
            # init matrices
            matrix = np.ones((nof_limits, nof_bins)) * big_M
            argmatrix = np.ones((nof_limits, nof_bins)) * np.nan

            # init first bin_index
            bin_index = 0
            for lim_index in range(nof_limits):
                matrix[lim_index, bin_index] = self._cost_of_move(
                    bin_index, lim_index, max_nof_bins, discount
                )

            # for all other bins
            for bin_index in range(1, max_nof_bins):
                for lim_index_next in range(max_nof_bins + 1):
                    # find best solution
                    tmp = []
                    for lim_index_before in range(max_nof_bins + 1):
                        tmp.append(
                            matrix[lim_index_before, bin_index - 1]
                            + self._cost_of_move(
                                lim_index_before, lim_index_next, max_nof_bins, discount
                            )
                        )
                    # store best solution
                    matrix[lim_index_next, bin_index] = np.min(tmp)
                    argmatrix[lim_index_next, bin_index] = np.argmin(tmp)

            # find indices
            self.method_outputs = {"matrix": matrix, "argmatrix": argmatrix}
            self.limits, _ = self._argmatrix_to_limits(max_nof_bins)
            self.method_outputs["limits"] = self.limits
        return self.limits


class Fixed(BinBase):
    def __init__(self, data, data_effect, feature, axis_limits):
        super(Fixed, self).__init__(feature, data, data_effect, axis_limits)

    def find(self, nof_bins: int, min_points: typing.Union[None, int] = 10, cat_limit: int = 10):
        self.method_args = {"nof_bins": nof_bins, "min_points": min_points, "cat_limit": cat_limit}

        if self._none_valid_binning():
            self.limits = False

        if self._is_categorical(cat_limit):
            return self.limits

        limits, dx = np.linspace(self.xs_min, self.xs_max, num=nof_bins + 1, endpoint=True, retstep=True)
        if min_points is not None:
            limits_valid = all(self._bin_valid(limits[i], limits[i + 1]) for i in range(nof_bins))
            self.limits = limits if limits_valid else False
        else:
            self.limits = limits

        return self.limits





from effector.global_effect_ale import ALE, RHALE
from effector.global_effect_pdp import PDP, DerivativePDP
from effector.global_effect_shap import SHAPDependence

from effector.regional_effect_ale import RegionalALE, RegionalRHALE
from effector.regional_effect_pdp import RegionalPDP, RegionalDerivativePDP
from effector.regional_effect_shap import RegionalSHAP

from effector.binning_methods import DynamicProgramming, Greedy, Fixed


import numpy as np
import effector

def test_gam():
    """
    Test the vectorized version of the SHAP function for a square model
    """

    np.random.seed(21)

    N = 1000
    T = 1000
    atol = 0.1
    rtol = 0.1

    data = np.stack([
            np.random.rand(N + 1),
            np.random.rand(N + 1),
        ], axis=1)

    model = lambda x: x[:, 0]**3/5 + x[:, 1] ** 2/5
    model_jac = lambda x: np.stack([3 * x[:, 0]**2/5, 2 * x[:, 1]/5], axis=1)

    x = np.linspace(0, 1, T)
    gt = {"x1": x**3/5, "x2": x**2/5, "heterogeneity": np.zeros_like(x), "x1_der": 3 * x**2/5, "x2_der": 2 * x/5}

    # Define test cases
    test_cases = [
        {"method": effector.PDP, "kwargs": {}},
        {"method": effector.DerivativePDP, "kwargs": {"model_jac": None}},
        {"method": effector.DerivativePDP, "kwargs": {"model_jac": model_jac}},
        {"method": effector.SHAPDependence, "kwargs": {}},
        {"method": effector.ALE, "kwargs": {}},
        {"method": effector.RHALE, "kwargs": {"model_jac": None}},
        {"method": effector.RHALE, "kwargs": {"model_jac": model_jac}}
    ]

    for test_case in test_cases:
        effector_class = test_case["method"]
        kwargs = test_case["kwargs"]

        # Instantiate the effector class
        eff = effector_class(data, model, nof_instances=100, **kwargs)

        for feature in [0, 1]:
            # Evaluate the effector and retrieve results
            y, heterogeneity = eff.eval(feature=feature, xs=x, heterogeneity=True, centering="zero_start")

            np.allclose(heterogeneity, gt["heterogeneity"], atol=atol, rtol=rtol)

            if effector_class not in [effector.DerivativePDP]:
                if feature == 0:
                    np.allclose(y, gt["x1"], atol=atol, rtol=rtol)
                elif feature == 1:
                    np.allclose(y, gt["x2"], atol=atol, rtol=rtol)
            else:
                if feature == 0:
                    np.allclose(y, gt["x1_der"], atol=atol, rtol=rtol)
                elif feature == 1:
                    np.allclose(y, gt["x2_der"], atol=atol, rtol=rtol)



import typing
import numpy as np
import copy
from effector import helpers

BIG_M = helpers.BIG_M
EPS = helpers.EPS


def compute_local_effects(
    data: np.ndarray, model: typing.Callable, limits: np.ndarray, feature: int
) -> np.ndarray:
    """Compute the local effects, permuting the feature of interest using the bin limits.

    Notes:
        The function (a) allocates the points in the bins based on the feature of interest (foi)
        and (b) computes the effect as the difference when evaluating the output setting the foi at the right and the
        left limit of the bin.

        Given that the bins are defined as a list [l_0, l_1, ..., l_k], and x_s of the i-th point belongs to the k-th bin:

        $$
        {df \over dx_s}(x^i) = {f(x_0^i, ... ,x_s=l_k, ..., x_D^i) - f(x_0^i, ... ,x_s=l_{k-1}, ..., x_D^i)
         \over l_k - l_{k-1}}
        $$


    Examples:
        >>> data = np.array([[1, 2], [2, 3.0]])
        >>> model = lambda x: np.sum(x, axis=1)
        >>> limits = np.array([1.0, 2.0])
        >>> data_effect = compute_local_effects(data, model, limits, feature=0)
        >>> data_effect
        array([1., 1.])

    Args:
        data: The training set, (N, D)
        model: The black-box model ((N, D) -> (N))
        limits: The bin limits, (K+1)
        feature: Index of the feature-of-interest

    Returns:
        data_effect: The local effect of each data point, (N)

    """
    assert data.ndim == 2

    # check that limits cover all data points
    assert limits[0] <= np.min(data[:, feature])
    assert limits[-1] >= np.max(data[:, feature])

    # for each point, find the bin-index it belongs to
    limits[-1] += EPS
    ind = np.digitize(data[:, feature], limits)
    assert np.alltrue(ind > 0)

    # compute effect
    right_lim = copy.deepcopy(data)
    left_lim = copy.deepcopy(data)
    right_lim[:, feature] = limits[ind]
    left_lim[:, feature] = limits[ind - 1]
    dx = limits[1] - limits[0]
    data_effect = model(right_lim) - model(left_lim)
    return np.squeeze(data_effect) / dx


def filter_points_in_bin(
    xs: np.ndarray, df_dxs: typing.Union[np.ndarray, None], limits: np.ndarray
) -> typing.Tuple[np.ndarray, typing.Union[np.ndarray, None]]:
    """
    Filter the points inside the bin defined by the `limits`.

    Notes:
        Filtering depends on whether `xs` lies in the interval [limits[0], limits[1]], not `df_dxs`.

    Examples:
        >>> xs = np.array([1, 2, 3])
        >>> df_dxs = np.array([32, 34, 36])
        >>> limits = np.array([1, 2])
        >>> xs, df_dxs = filter_points_in_bin(xs, df_dxs, limits)
        >>> xs
        array([1, 2])
        >>> df_dxs
        array([32, 34])

    Args:
        xs: The instances, (N)
        df_dxs: The instance-effects (N) or None
        limits: [Start, Stop] of the bin

    Returns:
        data: The instances in the bin, (nof_points_in_bin, D)
        data_effect: The instance-effects in the bin, (nof_points_in_bin, D) or None

    """
    filt = np.logical_and(limits[0] <= xs, xs <= limits[1])

    # return data
    xs = xs[filt]

    # return data effect if not None
    if df_dxs is not None:
        df_dxs = df_dxs[filt]
    return xs, df_dxs


def compute_bin_effect(
    xs: np.ndarray, df_dxs: np.ndarray, limits: np.ndarray
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Compute the mean effect in each bin.

    Notes:
        The function (a) allocates the instances in the bins and (b) aggregates the instance-level effects to compute
        the average bin-effect. If no instances lie in a bin, then the bin effect is NaN.

        $$
        \mathtt{bin\_effect}_k = {1 \over |i \in bin_k|} \sum_{i \in bin_k} \mathtt{effect}_i
        $$

    Examples:
        >>> n = 100
        >>> xs = np.ones([n]) - 0.5
        >>> df_dxs = np.ones_like(xs) * 10
        >>> limits = np.array([0., 1., 2.0])
        >>> bin_effects, ppb = compute_bin_effect(xs, df_dxs, limits)
        >>> bin_effects
        array([10., nan])
        >>> ppb
        array([100,   0])

    Parameters:
        xs: The s-th feature of the instances, (N)
        df_dxs: The effect wrt the s-th feature for each instance, (N)
        limits: The bin limits, (K+1)

    Returns:
        bin_effects: The average effect per bin, (K)
        points_per_bin: The number of points per bin, (K)
    """
    empty_symbol = np.NaN

    # find bin-index of points
    limits_enh = copy.deepcopy(limits).astype(float)
    limits_enh[-1] += EPS
    ind = np.digitize(xs, limits_enh)
    # assert np.alltrue(ind > 0)

    # bin effect is the mean of all points that lie in the bin
    nof_bins = limits.shape[0] - 1
    aggregated_effect = np.bincount(ind - 1, df_dxs, minlength=nof_bins)
    points_per_bin = np.bincount(ind - 1, minlength=nof_bins)

    # if no point lies in a bin, store Nan
    bin_effect_mean = np.divide(
        aggregated_effect,
        points_per_bin,
        out=np.ones(aggregated_effect.shape, dtype=float) * empty_symbol,
        where=points_per_bin != 0,
    )
    return bin_effect_mean, points_per_bin


def compute_bin_variance(
    xs: np.ndarray, df_dxs: np.ndarray, limits: np.ndarray, bin_effect_mean: np.ndarray
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """
    Compute the variance of the effect in each bin.

    Notes:
        The function (a) allocates the points in the bins and (b) computes the variance and the variance/nof points.
        If less than two points in a bin, NaN is passed.

        $$
        \mathtt{bin\_variance}_k = {1 \over |i \in bin_k|} \sum_{i \in bin_k}
        (\mathtt{effect}_i - \mathtt{bin\_effect}_k)^2
        $$

        $$
        \mathtt{bin\_estimator\_variance_k} = {\mathtt{bin\_variance}_k \over |i \in bin_k|}
        $$

    Examples:
        >>> n = 100
        >>> xs = np.ones([n]) - 0.5
        >>> df_dxs = np.ones_like(xs) * 10
        >>> limits = np.array([0., 1., 2.0])
        >>> bin_effect_mean, ppb = compute_bin_effect(xs, df_dxs, limits)
        >>> bin_variance, bin_estimator_variance = compute_bin_variance(xs, df_dxs, limits, bin_effect_mean)
        >>> bin_variance
        array([ 0., nan])
        >>> bin_estimator_variance
        array([ 0., nan])

        >>> xs = np.ones(4) * 0.5
        >>> df_dxs = np.array([1.0, 3.0, 3.0, 5.0])
        >>> limits = np.array([0, 1, 2.0])
        >>> bin_effect_mean = np.array([np.mean(df_dxs), np.NaN])
        >>> compute_bin_variance(xs, df_dxs, limits, bin_effect_mean)
        (array([ 2., nan]), array([0.5, nan]))

    Parameters:
        xs: The points we evaluate, (N)
        df_dxs: The effect of each point, (N, )
        limits: The bin limits (K+1)
        bin_effect_mean: Mean effect in each bin, (K)

    Returns:
        bin_variance: The variance in each bin, (K, )
        bin_estimator_variance: The variance of the estimator in each bin, (K, )

    """
    empty_symbol = np.NaN

    # find bin-index of points
    eps = 1e-8
    limits_enh = copy.deepcopy(limits).astype(float)
    limits_enh[-1] += eps
    ind = np.digitize(xs, limits_enh)
    # assert np.alltrue(ind > 0)

    # variance of the effect in each bin
    variance_per_point = (df_dxs - bin_effect_mean[ind - 1]) ** 2
    nof_bins = limits.shape[0] - 1
    aggregated_variance_per_bin = np.bincount(
        ind - 1, variance_per_point, minlength=nof_bins
    )
    points_per_bin = np.bincount(ind - 1, minlength=nof_bins)

    # if less than two points in a bin, store Nan
    bin_variance = np.divide(
        aggregated_variance_per_bin,
        points_per_bin,
        out=np.ones(aggregated_variance_per_bin.shape, dtype=float) * empty_symbol,
        where=points_per_bin > 1,
    )

    # the variance of the estimator
    bin_estimator_variance = np.divide(
        bin_variance,
        points_per_bin,
        out=np.ones(aggregated_variance_per_bin.shape, dtype=float) * empty_symbol,
        where=points_per_bin > 1,
    )
    return bin_variance, bin_estimator_variance


def fill_nans(x: np.ndarray) -> np.ndarray:
    """Replace NaNs with interpolated values.

    Examples:
        >>> x = np.array([1.0, np.NaN, 2.0])
        >>> fill_nans(x)
        array([1. , 1.5, 2. ])

        >>> x = np.array([1.0, np.NaN, np.NaN, np.NaN, 2.0])
        >>> fill_nans(x)
        array([1.  , 1.25, 1.5 , 1.75, 2.  ])

        >>> x = np.array([0.5, 1.0, np.NaN, np.NaN, np.NaN])
        >>> fill_nans(x)
        array([0.5, 1. , 1. , 1. , 1. ])

    Parameters:
        x: Time-series with NaNs, (T)

    Returns:
        x: Time-series values without NaNs, (T)
    """
    bin_effect_1 = copy.deepcopy(x)

    def nan_helper(y):
        return np.isnan(y), lambda z: z.nonzero()[0]

    nans, x = nan_helper(bin_effect_1)
    bin_effect_1[nans] = np.interp(x(nans), x(~nans), bin_effect_1[~nans])
    return bin_effect_1


def compute_accumulated_effect(
    x: np.ndarray,
    limits: np.ndarray,
    bin_effect: np.ndarray,
    dx: np.ndarray,
    square: bool = False,
) -> np.ndarray:
    """Compute the accumulated effect at each point `x`.

    Notes:
        The function implements the following formula:

        $$
        \mathtt{dx}[i] = \mathtt{limits}[i+1] - \mathtt{limits}[i]
        $$

        $$
        \mathtt{full\_bin\_acc} = \sum_{i=0}^{k_x - 1} \mathtt{dx}[i] * \mathtt{bin\_effect}[i]
        $$

        $$
        \mathtt{remainder} = (x - \mathtt{limits}[k_x-1])* \mathtt{bin\_effect}[k_x]
        $$

        $$
        f(x) =  \mathtt{full\_bin\_acc} + \mathtt{remainder}
        $$

    Notes:
        if `square=True`, then the formula is:
        $$
        \mathtt{full\_bin\_acc} = \sum_{i=0}^{k_x - 1} \mathtt{dx}^2[i] * \mathtt{bin\_effect}[i]
        $$

        $$
        \mathtt{remainder} = (x - \mathtt{limits}[k_x-1])^2* \mathtt{bin\_effect}[k_x]
        $$

    Examples:
        >>> x = np.array([-1.0, -0.5, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0])
        >>> limits = np.array([0, 1.5, 2.0])
        >>> bin_effect = np.array([1.0, -1.0])
        >>> dx = np.array([1.5, 0.5])
        >>> compute_accumulated_effect(x, limits, bin_effect, dx)
        array([0. , 0. , 0. , 0.5, 1. , 1.5, 1. , 1. , 1. ])

        >>> x = np.array([-1.0, -0.5, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0])
        >>> limits = np.array([0, 1.5, 2.0])
        >>> bin_effect = np.array([1.0, 1.0])
        >>> dx = np.array([1.5, 0.5])
        >>> compute_accumulated_effect(x, limits, bin_effect, dx)
        array([0. , 0. , 0. , 0.5, 1. , 1.5, 2. , 2. , 2. ])



    Parameters:
        x: The points we want to evaluate at, (T)
        limits: The bin limits, (K+1)
        bin_effect: The effect in each bin, (K)
        dx: The bin-widths, (K)
        square: Whether to square the width. If true, the effect is bin_effect * dx^2, otherwise bin_effect * dx

    Returns:
        y: The accumulated effect at each point, (T)


    """
    # find where each point belongs to
    ind = np.digitize(x, limits)

    # for each point, find the accumulated full-bin effect
    x_cumsum = (bin_effect * dx**2).cumsum() if square else (bin_effect * dx).cumsum()
    tmp = np.concatenate([[0, 0], x_cumsum])
    full_bin_effect = tmp[ind]

    # for each point, find the remaining effect
    tmp = np.concatenate([[limits[0]], limits[:-1], [BIG_M]])
    deltas = x - tmp[ind]
    deltas[deltas < 0] = 0  # if xs < left_limit => delta = 0
    deltas = deltas**2 if square else deltas
    tmp = np.concatenate([[0.0], bin_effect, [bin_effect[-1]]])
    remaining_effect = deltas * tmp[ind]

    # final effect
    y = full_bin_effect + remaining_effect
    return y


def compute_ale_params(xs: np.ndarray, df_dxs: np.ndarray, limits: np.ndarray) -> dict:
    """
    Compute all important parameters for the ALE plot.

    Examples:
        >>> # Example without interpolation
        >>> xs = np.array([0.5, 1.2, 2, 2.3])
        >>> df_dxs = np.array([30, 34, 15, 17])
        >>> limits = np.array([0, 1.5, 3.])
        >>> compute_ale_params(xs, df_dxs, limits)
        {'limits': array([0. , 1.5, 3. ]), 'dx': array([1.5, 1.5]), 'points_per_bin': array([2, 2]), 'bin_effect': array([32., 16.]), 'bin_variance': array([4., 1.]), 'bin_estimator_variance': array([2. , 0.5])}

        >>> # Example with interpolation
        >>> xs = np.array([1, 2, 2.8, 4])
        >>> df_dxs = np.array([31, 34, 37, 40])
        >>> limits = np.array([1, 3, 4])
        >>> compute_ale_params(xs, df_dxs, limits)
        {'limits': array([1, 3, 4]), 'dx': array([2, 1]), 'points_per_bin': array([3, 1]), 'bin_effect': array([34., 40.]), 'bin_variance': array([6., 6.]), 'bin_estimator_variance': array([2., 2.])}

    Args:
        xs: The values of s-th feature, (N)
        df_dxs: The effect wrt the s-th feature, (N)
        limits: The bin limits, (K+1)

    Returns:
        parameters: dict

    """
    # compute bin-widths
    dx = np.array([limits[i + 1] - limits[i] for i in range(len(limits) - 1)])

    # compute mean effect on each bin
    bin_effect_nans, points_per_bin = compute_bin_effect(xs, df_dxs, limits)

    # compute effect variance in each bin
    bin_variance_nans, bin_estimator_variance_nans = compute_bin_variance(
        xs, df_dxs, limits, bin_effect_nans
    )

    # interpolate NaNs
    bin_effect = fill_nans(bin_effect_nans)
    bin_variance = fill_nans(bin_variance_nans)
    bin_estimator_variance = fill_nans(bin_estimator_variance_nans)

    parameters = {
        "limits": limits,
        "dx": dx,
        "points_per_bin": points_per_bin,
        "bin_effect": bin_effect,
        "bin_variance": bin_variance,
        "bin_estimator_variance": bin_estimator_variance,
    }
    return parameters


def get_feature_types(data: np.ndarray, categorical_limit: int = 10) -> typing.List[str]:
    """Determine the type of each feature.

    Notes:
        A feature is considered as categorical if it has less than `cat_limit` unique values.

    Args:
        data: The dataset, (N, D)
        categorical_limit: Maximum unique values for a feature to be considered as categorical


    Returns:
        types: A list of strings, where each string is either `"cat"` or `"cont"`

    """

    types = [
        "cat" if len(np.unique(data[:, f])) < categorical_limit else "cont"
        for f in range(data.shape[1])
    ]
    return types


def compute_jacobian_numerically(
    model: typing.Callable, data: np.ndarray, eps: float = 1e-8
) -> np.ndarray:
    """Compute the Jacobian of the model using finite differences.

    Notes:
        The function computes the Jacobian of the model using finite differences. The formula is:

        $$
        \mathtt{J} = {\mathtt{model}(x + \mathtt{eps}) - \mathtt{model}(x) \over \mathtt{eps}}
        $$

    Examples:
        >>> data = np.array([[1, 2], [2, 3.0]])
        >>> model = lambda x: np.sum(x, axis=1)
        >>> compute_jacobian_numerically(model, data)
        array([[1., 1.],
               [1., 1.]])

    Args:
        data: The dataset, (N, D)
        model: The black-box model ((N, D) -> (N))
        eps: The finite difference step

    Returns:
        jacobian: The Jacobian of the model, (N, D)

    """
    assert data.ndim == 2
    jacobian = np.zeros_like(data)
    for f in range(data.shape[1]):
        data_plus = copy.deepcopy(data)
        data_plus[:, f] += eps
        jacobian[:, f] = (model(data_plus) - model(data)) / eps
    return jacobian



import numpy as np
import effector.binning_methods as binning_methods
import effector.helpers as helpers
import effector.utils as utils
from effector.partitioning import Regions, Tree
from effector.global_effect_ale import RHALE, ALE
from effector.global_effect_pdp import PDP, DerivativePDP
from effector.global_effect_shap import SHAPDependence
import typing
from typing import Callable, Optional, Union, List
from tqdm import tqdm
import copy

BIG_M = helpers.BIG_M


class RegionalEffectBase:
    empty_symbol = helpers.EMPTY_SYMBOL

    def __init__(
        self,
        method_name: str,
        data: np.ndarray,
        model: Callable,
        model_jac: Optional[Callable] = None,
        data_effect: Optional[np.ndarray] = None,
        nof_instances: Union[int, str] = 100,
        axis_limits: Optional[np.ndarray] = None,
        feature_types: Optional[List] = None,
        cat_limit: Optional[int] = 10,
        feature_names: Optional[List] = None,
        target_name: Optional[str] = None,
    ) -> None:
        """
        Constructor for the RegionalEffect class.
        """
        self.method_name = method_name.lower()
        self.model = model
        self.model_jac = model_jac

        # select nof_instances from the data
        self.nof_instances, self.indices = helpers.prep_nof_instances(
            nof_instances, data.shape[0]
        )
        self.data = data[self.indices, :]
        self.instance_effects = data_effect[self.indices, :] if data_effect is not None else None
        self.dim = self.data.shape[1]

        # set axis_limits
        axis_limits = (
            helpers.axis_limits_from_data(data) if axis_limits is None else axis_limits
        )
        self.axis_limits: np.ndarray = axis_limits

        # set feature types
        self.cat_limit = cat_limit
        feature_types = (
            utils.get_feature_types(data, cat_limit)
            if feature_types is None
            else feature_types
        )
        self.feature_types: list = feature_types

        # set feature names
        feature_names: list[str] = (
            helpers.get_feature_names(axis_limits.shape[1])
            if feature_names is None
            else feature_names
        )
        self.feature_names: list = feature_names

        # set target name
        self.target_name = "y" if target_name is None else target_name

        # state variables
        self.is_fitted: np.ndarray = np.ones([self.dim]) < 0

        # parameters used when fitting the regional effect
        self.method_args: typing.Dict = {}

        # dictionary with all the information required for plotting or evaluating the regional effects
        self.partitioners: typing.Dict[str, Regions] = {}
        self.tree_full: typing.Dict[str, Tree] = {}
        self.tree_pruned: typing.Dict[str, Tree] = {}
        self.tree_full_scaled: typing.Dict[str, Tree] = {}
        self.tree_pruned_scaled: typing.Dict[str, Tree] = {}

    def _fit_feature(
        self,
        feature: int,
        heter_func: Callable,
        heter_pcg_drop_thres: float = 0.1,
        heter_small_enough: float = 0.1,
        max_split_levels: int = 2,
        candidate_positions_for_numerical: int = 20,
        min_points_per_subregion: int = 10,
        candidate_foc: Union[str, List] = "all",
        split_categorical_features: bool = False,
    ):
        """
        Find the subregions for a single feature.
        """
        # init Region Extractor
        regions = Regions(
            feature,
            heter_func,
            self.data,
            self.instance_effects,
            self.feature_types,
            self.feature_names,
            self.target_name,
            self.cat_limit,
            candidate_foc,
            min_points_per_subregion,
            candidate_positions_for_numerical,
            max_split_levels,
            heter_pcg_drop_thres,
            heter_small_enough,
            split_categorical_features,
        )

        # apply partitioning
        regions.search_all_splits()
        regions.choose_important_splits()
        self.tree_full["feature_{}".format(feature)] = regions.splits_to_tree()
        self.tree_pruned["feature_{}".format(feature)] = regions.splits_to_tree(True)

        # store the partitioning object
        self.partitioners["feature_{}".format(feature)] = regions

        # update state
        self.is_fitted[feature] = True

    def refit(self, feature):
        if not self.is_fitted[feature]:
            self.fit(feature)

    def get_node_info(self, feature, node_idx):
        assert self.is_fitted[feature], "Feature {} has not been fitted yet".format(feature)
        assert self.tree_pruned["feature_{}".format(feature)] is not None, "Feature {} has no splits".format(feature)

        if self.tree_pruned_scaled is not None and "feature_{}".format(feature) in self.tree_pruned_scaled.keys():
            tree = self.tree_pruned_scaled["feature_{}".format(feature)]
        else:
            tree = self.tree_pruned["feature_{}".format(feature)]

        # assert node id exists
        assert node_idx in [node.idx for node in tree.nodes], "Node {} does not exist".format(node_idx)

        # find the node
        node = [node for node in tree.nodes if node.idx == node_idx][0]

        # get data
        data = node.data["data"]
        data_effect = node.data["data_effect"]
        name = node.name
        return data, data_effect, name

    def _create_fe_object(self, data, data_effect, feature_names):
        if self.method_name == "rhale":
            return RHALE(data, self.model, self.model_jac, data_effect=data_effect, feature_names=feature_names, target_name=self.target_name)
        elif self.method_name == "ale":
            return ALE(data, self.model, feature_names=feature_names, target_name=self.target_name)
        elif self.method_name == "shap":
            return SHAPDependence(data, self.model, feature_names=feature_names, target_name=self.target_name)
        elif self.method_name == "pdp":
            return PDP(data, self.model, feature_names=feature_names, target_name=self.target_name)
        elif self.method_name == "d-pdp":
            return DerivativePDP(data, self.model, self.model_jac, feature_names=feature_names, target_name=self.target_name)
        else:
            raise NotImplementedError

    def eval(self, feature, node_idx, xs, heterogeneity=False, centering=False):
        """
        Evaluate the regional effect for a given feature and node.

        Args:
            feature: the feature to evaluate
            node_idx: the node corresponding to the subregion to evaluate
            xs: the points at which to evaluate the regional effect
            heterogeneity: whether to return the heterogeneity.

                  - if `heterogeneity=False`, the function returns the mean effect at the given `xs`
                  - If `heterogeneity=True`, the function returns `(y, std)` where `y` is the mean effect and `std` is the standard deviation of the mean effect

            centering: whether to center the regional effect. The following options are available:

                - If `centering` is `False`, the regional effect is not centered
                - If `centering` is `True` or `zero_integral`, the regional effect is centered around the `y` axis.
                - If `centering` is `zero_start`, the regional effect starts from `y=0`.

        Returns:
            the mean effect `y`, if `heterogeneity=False` (default) or a tuple `(y, std)` otherwise

        """
        self.refit(feature)
        centering = helpers.prep_centering(centering)
        data, data_effect, _ = self.get_node_info(feature, node_idx)
        fe_method = self._create_fe_object(data, data_effect, None)
        return fe_method.eval(feature, xs, heterogeneity, centering)

    def fit(self, *args, **kwargs):
        raise NotImplementedError

    def plot(self,
             feature,
             node_idx,
             heterogeneity=False,
             centering=False,
             scale_x_list=None,
             scale_y=None,
             y_limits=None):

        self.refit(feature)

        if scale_x_list is not None:
            self.tree_full_scaled["feature_{}".format(feature)] = self.partitioners["feature_{}".format(feature)].splits_to_tree(False, scale_x_list)
            self.tree_pruned_scaled["feature_{}".format(feature)] = self.partitioners["feature_{}".format(feature)].splits_to_tree(True, scale_x_list)

        data, data_effect, name = self.get_node_info(feature, node_idx)
        feature_names = copy.deepcopy(self.feature_names)
        feature_names[feature] = name
        fe_method = self._create_fe_object(data, data_effect, feature_names)

        return fe_method.plot(
            feature=feature,
            heterogeneity=heterogeneity,
            centering=centering,
            scale_x=scale_x_list[feature] if scale_x_list is not None else None,
            scale_y=scale_y,
            y_limits=y_limits
            )

    def show_partitioning(self, features, only_important=True, scale_x_list=None):
        features = helpers.prep_features(features, self.dim)

        for feat in features:
            self.refit(feat)

            if scale_x_list is not None:
                tree_full_scaled = self.partitioners["feature_{}".format(feat)].splits_to_tree(True, scale_x_list)
                tree_pruned_scaled = self.partitioners["feature_{}".format(feat)].splits_to_tree(False, scale_x_list)
                tree_dict = tree_full_scaled if only_important else tree_pruned_scaled
            else:
                tree_dict = self.tree_pruned["feature_{}".format(feat)] if only_important else self.tree_full["feature_{}".format(feat)]

            print("Feature {} - Full partition tree:".format(feat))

            if tree_dict is None:
                print("No splits found for feature {}".format(feat))
            else:
                tree_dict.show_full_tree()

            print("-" * 50)
            print("Feature {} - Statistics per tree level:".format(feat))

            if tree_dict is None:
                print("No splits found for feature {}".format(feat))
            else:
                tree_dict.show_level_stats()

    def describe_subregions(
        self,
        features,
        only_important=True,
        scale_x_list: typing.Union[None, typing.List[dict]] = None,
    ):
        features = helpers.prep_features(features, self.dim)
        for feature in features:
            self.refit(feature)

            # it means it a categorical feature
            if self.tree_full["feature_{}".format(feature)] is None:
                continue

            feature_name = self.feature_names[feature]
            if only_important:
                tree = self.tree_pruned["feature_{}".format(feature)]
                if len(tree.nodes) == 1:
                    print("No important splits found for feature {}".format(feature))
                    continue
                else:
                    print("Important splits for feature {}".format(feature_name))
            else:
                print("All splits for feature {}".format(feature_name))
                tree = self.tree_full["feature_{}".format(feature)]

            max_level = max([node.level for node in tree.nodes])
            for level in range(1, max_level+1):
                previous_level_nodes = tree.get_level_nodes(level-1)
                level_nodes = tree.get_level_nodes(level)
                type_of_split_feature = level_nodes[0].data["feature_type"]
                foc_name = self.feature_names[level_nodes[0].data["feature"]]
                print("- On feature {} ({})".format(foc_name, type_of_split_feature))

                position_split_formatted = (
                    "{:.2f}".format(level_nodes[0].data["position"])
                    if scale_x_list is None
                    else "{:.2f}".format(
                        level_nodes[0].data["position"] * scale_x_list[level_nodes[0].data["feature"]]["std"]
                        + scale_x_list[level_nodes[0].data["feature"]]["mean"]
                    )
                )
                print("  - Position of split: {}".format(position_split_formatted))

                weight_heter_before = np.sum([node.data["weight"] * node.data["heterogeneity"] for node in previous_level_nodes])
                print("  - Heterogeneity before split: {:.2f}".format(weight_heter_before))

                weight_heter = np.sum([node.data["weight"] * node.data["heterogeneity"] for node in level_nodes])
                print("  - Heterogeneity after split: {:.2f}".format(weight_heter))
                weight_heter_drop = weight_heter_before - weight_heter
                print("  - Heterogeneity drop: {:.2f} ({:.2f} %)".format(
                    weight_heter_drop, weight_heter_drop / weight_heter_before * 100)
                )

                nof_instances_before = [nod.data["nof_instances"] for nod in previous_level_nodes]
                print("  - Number of instances before split: {}".format(nof_instances_before))
                nof_instances = [nod.data["nof_instances"] for nod in level_nodes]
                print("  - Number of instances after split: {}".format(nof_instances))



import numpy as np
import logging
import typing
from effector import helpers
import effector
from effector import DerivativePDP, PDP
from effector.global_effect_pdp import pdp_1d_vectorized, pdp_1d_non_vectorized, pdp_nd_non_vectorized, pdp_nd_vectorized
import tqdm

class HIndex:
    """
    H-Index interaction measure, as described in Friedman and Popescu (2008), https://arxiv.org/abs/0811.1679
    """

    empty_symbol = 1e10

    def __init__(self, data, model, nof_instances=1000):
        """

        Args
        ---
            data (np.ndarray): The dataset, shape (nof_instances, nof_features)
            model (function): The model to explain, takes as input data and returns predictions
            nof_instances (int): The number of instances to use for the explanation
        """
        # setters
        self.nof_instances, self.indices = helpers.prep_nof_instances(nof_instances, data.shape[0])
        self.data = data[self.indices]
        self.model = model
        self.nof_features = self.data.shape[1]

        # init
        self.interaction_matrix = np.ones((self.nof_features, self.nof_features)) * self.empty_symbol
        self.one_vs_all_matrix = np.ones((self.nof_features)) * self.empty_symbol

        # flags
        self.fitted_interaction_matrix = False
        self.fitted_one_vs_all_matrix = False

    def fit(self, pairwise_matrix=True, one_vs_all=True):
        if pairwise_matrix:
            logging.info("\nH-Index interaction matrix: Start fitting")
            for i in range(self.nof_features):
                print("Feature: ", i)
                self.interaction_matrix[i, i] = 0
                for j in range(i + 1, self.nof_features):
                    self.eval_pairwise(i, j)
            self.fitted_interaction_matrix = True
            logging.info("H-Index interaction matrix: Fitting done")

        if one_vs_all:
            logging.info("\nH-Index one-vs-all matrix: Start fitting")
            for i in range(self.nof_features):
                print("Feature: ", i)
                self.eval_one_vs_all(i)
            self.fitted_one_vs_all_matrix = True
            logging.info("H-Index one-vs-all matrix: Fitting done")

    def plot(self, interaction_matrix=True, one_vs_all=True):
        # if not fitted, fit
        if not self.fitted_interaction_matrix and interaction_matrix:
            self.fit(pairwise_matrix=True, one_vs_all=False)
        if not self.fitted_one_vs_all_matrix and one_vs_all:
            self.fit(pairwise_matrix=False, one_vs_all=True)

        # plot
        if interaction_matrix:
            self._plot_interaction_matrix()
        if one_vs_all:
            self._plot_one_vs_all_matrix()

    def eval_pairwise(self, feat_1, feat_2):
        """Evaluate the interaction between feature 1 and feature 2

        Args
        ---
            feat_1 (int): The index of feature 1
            feat_2 (int): The index of feature 2

        Returns
        ---
            float: The interaction between feature 1 and feature 2
        """
        if self.interaction_matrix[feat_1, feat_2] != self.empty_symbol:
            return self.interaction_matrix[feat_1, feat_2]

        x1 = self.data[:, feat_1]
        x2 = self.data[:, feat_2]
        pdp_1 = self._pdp_1d(feat_1, x1)
        pdp_2 = self._pdp_1d(feat_2, x2)
        pdp_12 = self._pdp_2d(feat_1, feat_2, x1, x2)

        nom = np.mean(np.square(pdp_12 - pdp_1 - pdp_2))
        denom = np.mean(np.square(pdp_12))
        interaction = nom / denom

        self.interaction_matrix[feat_1, feat_2] = interaction
        self.interaction_matrix[feat_2, feat_1] = interaction
        return interaction

    def eval_one_vs_all(self, feat):
        if self.one_vs_all_matrix[feat] != self.empty_symbol:
            return self.one_vs_all_matrix[feat]

        pdp_1 = self._pdp_1d(feat, self.data[:, feat])
        pdp_minus_1 = self._pdp_minus_1d(feat, self.data[:, [i for i in range(self.nof_features) if i != feat]])

        yy = self.model(self.data)
        nom = np.mean(np.square(yy - pdp_minus_1 - pdp_1))
        denom = np.mean(np.square(yy))
        interaction = nom / denom

        self.one_vs_all_matrix[feat] = interaction
        return interaction

    def _pdp_1d(self, feature, x):
        # eval normalized pdp for feature at x
        yy = pdp_1d_vectorized(self.model, self.data, x, feature, uncertainty=False, model_returns_jac=False)
        c = np.mean(yy)
        return yy - c

    def _pdp_2d(self, feature1, feature2, x1, x2):
        features = [feature1, feature2]
        x = np.stack([x1, x2], axis=-1)
        yy = pdp_nd_vectorized(self.model, self.data, x, features, uncertainty=False, model_returns_jac=False)
        c = np.mean(yy)
        return yy - c

    def _pdp_minus_1d(self, feature, x):
        # features are all other features
        features = [i for i in range(self.nof_features) if i != feature]
        assert len(features) == x.shape[1]

        yy = pdp_nd_vectorized(self.model, self.data, x, features, uncertainty=False, model_returns_jac=False)
        c = np.mean(yy)
        return yy - c

    def _plot_interaction_matrix(self):
        import matplotlib.pyplot as plt
        plt.figure(figsize=(10, 10))
        plt.title("Interaction Matrix")
        # assign the mean to the diagonal values w
        mu = np.mean(self.interaction_matrix)
        for i in range(self.nof_features):
            self.interaction_matrix[i, i] = mu

        # plot as a heatmap
        plt.imshow(self.interaction_matrix)
        plt.colorbar()
        # show grid values
        for i in range(self.nof_features):
            for j in range(self.nof_features):
                plt.text(j, i, round(self.interaction_matrix[i, j], 2), ha="center", va="center", color="w")

        plt.xticks(np.arange(self.nof_features), labels=["feature {}".format(i+1) for i in range(self.nof_features)])
        plt.yticks(np.arange(self.nof_features), labels=["feature {}".format(i+1) for i in range(self.nof_features)])
        plt.show()

    def _plot_one_vs_all_matrix(self):
        import matplotlib.pyplot as plt
        # plot as horizontal bar chart
        plt.figure(figsize=(10, 10))
        plt.title("H-index for all features")
        plt.barh(np.arange(self.nof_features), self.one_vs_all_matrix)
        plt.yticks(np.arange(self.nof_features), labels=["feature {}".format(i+1) for i in range(self.nof_features)])
        plt.xlabel("H-index")
        plt.show()


class REPID:
    """REPID definition of interaction between feature x_j and all the others, i.e. the variance of the dPDP plot
    """
    empty_symbol = 1e10

    def __init__(self, data, model, model_jac, nof_instances):
        self.nof_instances, self.indices = helpers.prep_nof_instances(nof_instances, data.shape[0])
        self.data = data[self.indices]
        self.dim = self.data.shape[1]

        self.model = model
        self.model_jac = model_jac
        self.nof_instances = nof_instances
        self.nof_features = data.shape[1]

        self.interaction_matrix = np.ones((self.nof_features, self.nof_features)) * self.empty_symbol
        self.one_vs_all_matrix = np.ones(self.nof_features) * self.empty_symbol

        self.fitted_interaction_matrix = False
        self.fitted_one_vs_all_matrix = False

    def fit(self, features: typing.Union[int, str, list] = "all",):
        features = helpers.prep_features(features, self.dim)
        for i, feat in enumerate(features):
            print("Feature: ", feat)
            self.eval_one_vs_all(feat)
        self.fitted_one_vs_all_matrix = True

    def eval_one_vs_all(self, feat):
        if self.one_vs_all_matrix[feat] != self.empty_symbol:
            return self.one_vs_all_matrix[feat]

        axis_limits = effector.helpers.axis_limits_from_data(self.data)
        pdp = PDP(self.data, self.model, axis_limits)
        pdp.fit(feat)

        # find the interaction index
        start = axis_limits[:, feat][0]
        stop = axis_limits[:, feat][1]
        x = np.linspace(start, stop, 21)
        x = 0.5 * (x[:-1] + x[1:])
        mu, std, stderr = pdp.eval(feature=feat, xs=x, uncertainty=True, centering="zero_start") # "zero_integral")
        interaction = np.mean(std)

        # store the interaction index
        self.one_vs_all_matrix[feat] = interaction
        return interaction

    def plot(self):
        if self.fitted_one_vs_all_matrix is False:
            self.fit()

        import matplotlib.pyplot as plt
        # plot as horizontal bar chart
        plt.figure(figsize=(10, 10))
        plt.title("REPID for all features")
        plt.barh(np.arange(self.nof_features), self.one_vs_all_matrix)
        plt.yticks(np.arange(self.nof_features), labels=["feature {}".format(i+1) for i in range(self.nof_features)])
        plt.xlabel("REPID index")
        plt.show()
